from __future__ import annotations

import anyio
import sys
import time
from contextlib import asynccontextmanager, nullcontext
from functools import partial
from pathlib import Path as FSPath
from tempfile import TemporaryDirectory

from moat.util import (
    # pylint:disable=no-name-in-module,
    CtxObj,
    NotGiven,
    ValueEvent,
    attrdict,
    combine_dict,
    merge,
)
from moat.lib.config import CFG
from moat.lib.path import P, Root
from moat.link.backend import get_backend
from moat.link.client import Link
from moat.link.server import Server

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.path import Path
    from moat.link.backend import Backend
    from moat.link.client import LinkCommon, LinkSender

    from collections.abc import AsyncIterator
    from typing import Literal, Never, Self


otm = time.time

_seq = 0


# TODO launch a "real" broker instead


async def run_broker(cfg, *, task_status):
    """
    Runs a basic MQTT broker.

    The task status returns the port we're listening on.
    """
    cfg  # pyright:ignore # noqa:B018

    if False:
        # Just use the system MQTT broker.
        # TODO: This requires cleaning up retained messages.
        task_status.started(1883)
        return

    elif False:
        # Use the experimental moat.lib.mqtt broker.
        from moat.lib.mqtt import AsyncMQTTBroker  # noqa:PLC0415

        broker = AsyncMQTTBroker(("127.0.0.1", 0))
        await broker.serve(task_status=task_status)

    else:
        # Use a standalone instance of FlashMQ.
        from anyio.pytest_plugin import FreePortFactory  # noqa: PLC0415
        from socket import SOCK_STREAM  # noqa: PLC0415

        port = FreePortFactory(SOCK_STREAM)()

        async with (
            anyio.TemporaryDirectory() as td,
            anyio.create_task_group() as tg,
        ):
            tf = anyio.Path(td) / "config"
            await tf.write_text(
                f"""\
allow_anonymous true
retained_messages_mode enabled_without_persistence
thread_count 1

log_level info
log_subscriptions true

listen {{
    protocol mqtt
    port {port}
    inet4_bind_address 127.0.0.1
}}
"""
            )

            tg.start_soon(
                partial(
                    anyio.run_process,
                    ["flashmq", "-c", str(tf)],
                    stderr=sys.stderr,
                    stdout=sys.stdout,
                    env=dict(HOME=td),
                )
            )
            for _ in range(20):
                try:
                    sock = await anyio.connect_tcp("127.0.0.1", port)
                except OSError:
                    await anyio.sleep(0.1)
                else:
                    await sock.aclose()
                    break
            else:
                raise RuntimeError("Could not connect to FlashMQ")

            task_status.started(port)


class Scaffold(CtxObj):
    """
    Basic testcase runner for testing with an ephemeral MQTT server.

    If @cfg is `True`, place our mod in (and use) the global `moat.link`
    subconfig. Otherwise, pass in the `moat` subconfig.
    """

    tempdir: FSPath | None

    def __init__(
        self, cfg: attrdict | Literal[True], use_servers=True, tempdir: str | None = None
    ):
        cf = attrdict(
            backend=attrdict(
                driver="mqtt",
                codec="std-cbor",
                keep_alive=9999,
            ),
            server=attrdict(
                ping=attrdict(
                    cycle=0.5,
                    gap=0.15,
                    override=True,
                ),
                timeout=attrdict(startup=3),
            ),
        )
        if cfg is True:
            CFG.mod(P("moat.link"), cf)
            self.cfg = CFG.result.moat.link
        else:
            self.cfg = cfg.link
            merge(self.cfg, cf)
        self._tempdir = tempdir

        if not use_servers:
            self.cfg.client.init_timeout = None

    @asynccontextmanager
    async def _ctx(self) -> AsyncIterator[Self]:
        Root.set(self.cfg.root)

        with (
            nullcontext(
                self._tempdir,
            )
            if self._tempdir is not None
            else TemporaryDirectory() as tempdir
        ):
            assert tempdir is not None
            self.tempdir = FSPath(tempdir)
            async with (
                anyio.create_task_group() as tg,
                anyio.create_task_group() as self.tg,
            ):
                bport = await tg.start(run_broker, self.cfg)
                self.cfg.backend.port = bport
                try:
                    yield self
                finally:
                    with anyio.CancelScope(shield=True):
                        await anyio.sleep(0.1)
                        self.tg.cancel_scope.cancel()  # pyright:ignore
                        await anyio.sleep(0.1)
                        tg.cancel_scope.cancel()  # pyright:ignore
                        await anyio.sleep(0.1)

    async def backend(self, cfg: dict | None = None, **kw):
        """
        Start a backend (background task)
        """
        return await self.tg.start(self._run_backend, cfg, kw)

    @asynccontextmanager
    async def backend_(self, cfg: dict | None, **kw) -> AsyncIterator[Backend]:
        """
        Start a backend (async context manager).
        """
        cfg = combine_dict(cfg, self.cfg, cls=attrdict) if cfg else self.cfg
        async with get_backend(cfg, **kw) as bk:
            yield bk

    async def _run_backend(self, cfg: dict | None, kw: dict, *, task_status) -> Backend:
        """
        Start a backend (Helper).
        """
        async with self.backend_(cfg, **kw) as bk:
            task_status.started(bk)
            await anyio.sleep_forever()
            raise AssertionError

    async def server(self, cfg: dict | None = None, **kw) -> tuple[Server, list[dict]]:
        """
        Start a server (background task)

        Returns the server object and the ports it runs on.
        """
        return await self.tg.start(self._run_server, cfg, kw)

    async def _run_server(self, cfg, kw, *, task_status) -> None:
        """
        Run a basic MoaT-Link server. (Helper task)
        """
        cfg = combine_dict(cfg, self.cfg, cls=attrdict) if cfg else self.cfg
        if "ports" in cfg["server"]:
            cfg["server"]["ports"]["main"]["port"] = 0
        cfg["server"]["port"] = 0
        if self.tempdir is not None:
            cfg["server"]["save"]["dir"] = self.tempdir / "data"

        name = kw.pop("name", None)
        if name is None:
            global _seq
            _seq += 1
            name = f"S_{_seq}"

        s = Server(cfg, name, **kw)
        await self.tg.start(s.serve)
        task_status.started(s)
        await anyio.sleep_forever()

    @asynccontextmanager
    async def server_(self, cfg: dict | None = None, **kw) -> AsyncIterator[Server]:
        """
        Runs a basic MoaT-Link server. (async context manager)
        """
        async with anyio.create_task_group() as tg:
            yield await tg.start(self._run_server, cfg, kw)
            tg.cancel_scope.cancel()

    async def client(self, *a, **kw):
        """
        Start a client (background task)
        """
        cl = await self.tg.start(partial(self._run_client, *a, **kw))
        return cl

    async def _run_client(self, *a, task_status, **kw) -> Never:
        async with self.client_(*a, **kw) as cl:
            task_status.started(cl)
            await anyio.sleep_forever()
        raise AssertionError("unreachable")

    @asynccontextmanager
    async def client_(
        self, cfg: dict | None = None, cli: LinkCommon | None = None, name=None
    ) -> AsyncIterator[LinkSender]:
        """
        Start a client (async context manager)
        """
        if cli is None:
            cfg = combine_dict(cfg, self.cfg, cls=attrdict) if cfg else self.cfg

            global _seq
            _seq += 1
            name = f"C_{_seq}"

            cli = Link(cfg, name)

        async with cli as li:
            yield li

    @asynccontextmanager
    async def do_watch(
        self, path: Path, exp=NotGiven, n: int = 0, **kw
    ) -> AsyncIterator[ValueEvent]:
        """
        Run a client that expects ``exp`` and appends all non-exp
        results to ``rd``.

        All other args+kw are forwarded to `Link.d_watch`.

        The results are appended to a list that's posted to the event when
        `do_watch` ends (for whatever reason).

        Args:
            path: The Link path to watch.
            exp: Expected element, or `NotGiven`.
            n: Exit after ``n`` messages. Default: Don't.
        """
        async with self.client_() as c, anyio.create_task_group() as tg:
            evt = ValueEvent()

            @tg.start_soon
            async def work():
                res = []
                try:
                    async with c.d_watch(path, **kw) as mon:
                        async for r in mon:
                            if kw.get("meta"):
                                t = time.time()
                                assert t - 1 < r[-1].timestamp < t
                            elif not kw.get("subtree"):
                                # neither meta nor subtree
                                r = (r,)  # noqa:PLW2901
                            if (
                                exp is not NotGiven
                                and (not kw.get("subtree") or not len(r[0]))
                                and r[kw.get("subtree", 0)] == exp
                            ):
                                # Skip if toplevel match
                                return
                            res.append(r)
                            if len(res) == n:
                                return
                finally:
                    evt.set(res)

            yield evt
            tg.cancel_scope.cancel()



"""
This module contains the basic MoaT-Link data model.
"""

from __future__ import annotations

from logging import getLogger

from attrs import define, field

from moat.util import (
    NotGiven,
    attrdict,
    combine_dict,
)
from moat.lib.path import (
    Path,
    PathLongener,
    PathShortener,
)
from moat.link.meta import MsgMeta
from moat.util.exc import ExpKeyError

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Key

    from collections.abc import Awaitable, Callable, Iterator
    from typing import Any, Self

__all__ = ["Node", "NodeFinder"]

logger = getLogger(__name__)


def _keys_repr(x):
    return ",".join(str(k) for k in x)


@define
class Node:
    """Represents one MoaT-Link item."""

    _data: Any = field(init=False, default=NotGiven)
    _meta: MsgMeta | None = field(init=False, default=None)

    _sub: dict[Key, Node] = field(init=False, factory=dict, repr=_keys_repr)  # sub-entries

    def set(self, item: Path, data: Any, meta: MsgMeta, force: bool = False) -> bool | None:
        """Save new data below this node.

        Return semantics:

        * If ``force`` is set:
          * ``True``: incoming timestamp is newer
          * ``None``: incoming data are equal
          * ``False``: otherwise
        * If ``force`` is not set:
          * ``None``: incoming data are equal
          * ``False``: incoming timestamp is older
          * ``True``: otherwise
        """
        assert isinstance(meta, MsgMeta)
        s = self.get(item)
        if s._meta is not None:  # noqa:SLF001
            same = s._data == data  # noqa:SLF001
            if force:
                if meta.timestamp > s._meta.timestamp:  # noqa:SLF001
                    s.set_(item, data, meta)
                    return True
                if same:
                    return None
                return False
            if same:
                return None
            if meta.timestamp < s._meta.timestamp:  # noqa:SLF001
                return False
        s.set_(item, data, meta)
        return True

    def set_(self, path: Path, data: Any, meta: MsgMeta):
        "Low-level node data setter. The (sub)path is not stored by default."
        path  # noqa:B018
        self._data = data
        self._meta = meta

    @property
    def data(self) -> Any:
        "return current data, raises ValueError if empty"
        if self._data is NotGiven:
            raise ValueError("empty node")
        return self._data

    @property
    def data_(self) -> Any:
        "return current data, returns NotGiven if empty"
        return self._data

    def keys(self):  # noqa: D102
        return self._sub.keys()

    def items(self):  # noqa: D102
        return self._sub.items()

    def __bool__(self) -> bool:
        "check if data exist"
        return self._data is not NotGiven

    def __eq__(self, other):
        if self._data != other._data:
            return False
        if self._sub != other._sub:
            return False
        return True

    def _dump_x(self):
        # Iterator that returns a serialization of this node tree.
        ps = PathShortener()
        for p, d, m in self._dump_x_(()):
            s, p = ps.short(p)  # noqa:PLW2901
            yield (s, p, d, *(m.dump() if m is not None else ()))

    def _dump_x_(self, path):
        # Helper for _dump_x
        if self._data is not NotGiven:
            yield path, self._data, self._meta
        for k, v in self._sub.items():
            yield from v._dump_x_(
                path + (k,),
            )

    def dump(self):
        """
        An iterator that returns a path-shortened serialization of this
        node tree.
        """
        # The naïve method (in `_dump_x`) creates a full-path tuple for
        # each node, all of which the PathShortener will throw away.
        #
        # This code yields the exact same data – without that overhead.
        # The old code is kept (a) because it's more easily understood,
        # (b) for unit testing.

        yield from self._dump((), 0)

    def _dump(self, path, level):
        if self._data is not NotGiven:
            yield (level, path, self._data, *(self._meta.dump() if self._meta is not None else ()))
            level += len(path)
            path = ()
        for k, v in self._sub.items():
            it = iter(v._dump(path + (k,), level))  # noqa: SLF001
            try:
                d = next(it)
            except StopIteration:
                pass
            else:
                yield d
                level += len(path)
                path = ()
                yield from it

    def load(self, force=False):
        """
        receives a data stream created by `dump`.

        if @force is set, overwrite existing data even if newer.
        """
        # TODO mirror dump() and do this without a PathLongener
        pl = PathLongener()
        while True:
            s, p, d, *m = yield
            m = MsgMeta.restore(m, NotGiven)
            p = pl.long(s, p)
            n = self.get(p)
            if force or n.meta is None or n.meta.timestamp < m.timestamp:
                n._data = d  # noqa: SLF001
                n._meta = m  # noqa: SLF001

    @property
    def meta(self) -> MsgMeta | None:
        "return current metadata"
        return self._meta

    @meta.deleter
    def meta(self) -> None:
        "Clear metadata"
        self._meta = None

    def __delitem__(self, item) -> None:
        """
        Remove an item. (It must be empty.)

        **Warning** Don't call this unless the timeout for deletion has passed.
        """
        d = self._sub[item]

        if d._sub or d.data_ is not NotGiven:  # noqa: SLF001
            raise ValueError(item)
        del self._sub[item]

    def __getitem__(self, item) -> Node:
        """Look up the entry.

        Raises KeyError if it doesn't exist.
        """
        if isinstance(item, Path):
            s = self
            for k in item:
                try:
                    s = s._sub[k]  # noqa:SLF001
                except KeyError:
                    raise ExpKeyError(k) from None
        else:
            s = self._sub[item]

        if s._data is NotGiven:  # noqa:SLF001
            raise ExpKeyError(item)
        return s

    def get(self, item, create=None) -> Node:
        """Look up an entry. Create if it doesn't exist.

        Unlike data[key], an "empty" key is not an error.
        """
        if item is Ellipsis:
            return self

        if isinstance(item, Path):
            s = self
            for n, k in enumerate(item):
                if isinstance(k, Path):
                    # import traceback
                    # logger.warning("Looking up %r\n%s", item, ''.join(traceback.format_stack()))
                    logger.warning("Looking up %r\n", item)
                    s = s.get(k, create=create)
                    continue
                try:
                    s = s._sub[k]  # noqa:SLF001
                except KeyError:
                    if create is False:
                        raise
                    s = s.add_child(k)
                else:
                    if create is True and n == len(item) - 1 and s._data is not NotGiven:  # noqa: SLF001
                        raise KeyError(k)
            return s

        try:
            res = self._sub[item]
        except KeyError:
            if create is False:
                raise
            return self.add_child(item)
        else:
            if create is True and res._data is not NotGiven:  # noqa: SLF001
                raise KeyError(item)
            return res

    def add_child(self, item):
        """Create and register a new child node.

        Override this method to return a different node type for
        specific child names.

        Args:
            item: the key for the new child.

        Returns:
            The newly-created child node.

        Raises:
            ValueError: if *item* already exists.
        """
        if item in self._sub:
            raise ValueError("exists")
        self._sub[item] = s = type(self)()
        return s

    def __iter__(self) -> Iterator[tuple[Key, Node]]:
        """
        Return a list of keys under this node.
        """
        return iter(self._sub.items())

    def __contains__(self, item) -> bool:
        if isinstance(item, Path):
            s = self
            for k in item:
                try:
                    s = s._sub[k]  # noqa:SLF001
                except KeyError:
                    return False
            return True

        return item in self._sub

    def deleted(self) -> bool:
        """
        Check whether this tick has been marked as deleted.
        """
        return self._data is NotGiven

    async def walk(
        self,
        proc: Callable[[Path, Node], Awaitable[bool | None]],
        max_depth: int = 999999,
        min_depth: int = 0,
        timestamp: float = 0,
        depth_first: bool = False,
        force: bool = False,
    ):
        """
        Calls coroutine ``proc(node,Subpath)`` on this node and all its children.

        Deleted nodes are passed if they still have a Meta entry.

        if @depth_first is not set and @proc explicitly returns False,
        the subtree is skipped.

        if @force is set, also visit empty nodes.
        """

        async def _walk(s, p):
            if depth_first and (max_depth is None or max_depth > len(p)):
                for k, v in s._sub.items():  # noqa: SLF001
                    await _walk(v, p / k)

            if (min_depth is None or min_depth <= len(p)) and (
                force or (s.meta is not None and s.meta.timestamp >= timestamp)
            ):
                if await proc(p, s) is False:
                    return

            if not depth_first and (max_depth is None or max_depth > len(p)):
                for k, v in list(s._sub.items()):  # noqa: SLF001
                    await _walk(v, p / k)

        await _walk(self, Path())

    def finder(self, path: Path) -> NodeFinder:
        """
        Find the destination node of a path, including wildcards.

        This node represents a search path.
        """
        return NodeFinder(self).at(path)

    def search(self, path: Path) -> Node:
        """
        Find the destination node of a path, including wildcards.

        This node represents the pattern.
        """
        return self.finder(path).result

    def collect(self, path: Path, keep: bool = False) -> dict:
        """
        Collate data from all matching branches for a path.

        This method combines all data nodes found in the matching
        ``NodeFinder`` branches, with more specific matches overriding
        less specific ones.
        """
        nf = NodeFinder(self)
        for p in path:
            try:
                nf.step(p)
            except KeyError:
                break

        res: dict = attrdict()
        for node in reversed(nf.matches):
            if node.data_ is not NotGiven:
                res = combine_dict(node.data_, res, cls=attrdict, keep=keep)

        return res


class NodeFinder:
    """A generic object that can walk down a possibly-wildcard-equipped path.

    Example: given a path ``one.two.three`` and a root with subtree ``*.three``,
    ``NodeFinder(root).step(one).step(two).step(three).result`` will return
    the node at ``*.three`` (assuming that nothing more specific hangs off
    the root).

    If nothing is found, raises `KeyError`.
    """

    def __init__(self, src):
        self.steps = ((src, 0, 0),)

    @staticmethod
    def _range_wildcards(node):
        """Yield `(min,max,node)` tuples for valid range wildcard children."""

        for key, sub in node._sub.items():  # noqa:SLF001
            if not isinstance(key, tuple) or len(key) != 2:
                continue
            n, m = key
            if type(n) is not int or type(m) is not int:
                continue
            if n < 1:
                continue
            if m != 0 and m < n:
                continue
            yield n, m, sub

    def copy(self) -> Self:
        """
        Create a clone with the current path.
        """
        res = object.__new__(NodeFinder)
        res.steps = self.steps
        return res

    def at(self, path: Path) -> Self:
        """Advance this finder by all elements in *path*."""
        for elem in path:
            self.step(elem)
        return self

    def step(self, name: str | int | bool | None, new=False) -> None:
        """
        Walk a single hierarchy step, observing wildcards. Note that ``#``
        means *one or more*, i.e. it will not match an empty path element.

        Args:
            name: the path element to look at.

        Don't use the *new* argument; it only exists for override compatibility.
        """
        if new:
            raise ValueError("I can't create new nodes.")

        steps = []
        for node, min_more, max_more in self.steps:
            if min_more == 0:
                if name in node:
                    steps.append((node.get(name), 0, 0))
                if "+" in node:
                    steps.append((node.get("+"), 0, 0))
                for n, m, sub in self._range_wildcards(node):
                    steps.append((sub, n - 1, None if m == 0 else m - 1))
                if "#" in node:
                    steps.append((node.get("#"), 0, None))

            if max_more is None or max_more > 0:
                steps.append((
                    node,
                    max(min_more - 1, 0),
                    None if max_more is None else max_more - 1,
                ))
            # Nodes found with '#' or range wildcards stay on the list
            # so that they can match additional path elements.
        if not steps:
            raise KeyError(name)
        self.steps = steps

    @property
    def result(self) -> Node:
        """Return the highest-priority matching node."""
        for node in self.matches:
            return node
        raise KeyError("No matching wildcard state")

    @property
    def matches(self) -> tuple[Node, ...]:
        """Return all applicable branch nodes in precedence order."""

        seen: set[int] = set()
        res: list[Node] = []
        for node, min_more, _max_more in self.steps:
            if min_more != 0:
                continue
            node_id = id(node)
            if node_id in seen:
                continue
            seen.add(node_id)
            res.append(node)
        return tuple(res)



"""
Runtime helper for code snippets stored in MoaT-Link data.
"""

from __future__ import annotations

import anyio
import ast
from anyio.to_thread import run_sync

from moat.util import NotGiven, combine_dict
from moat.lib.path import P, Path

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from types import CodeType

    from moat.link.client import LinkSender


class ReturnValue(BaseException):
    """
    Internal signal used to return values from compiled snippets.
    """

    value: Any

    def __init__(self, value: Any) -> None:
        super().__init__(value)
        self.value = value


class _ReturnRewriter(ast.NodeTransformer):
    """
    Rewrite module-level ``return`` statements to ``raise ReturnValue``.
    """

    _scope_level: int

    def __init__(self) -> None:
        super().__init__()
        self._scope_level = 0

    def _enter_scope(self, node: ast.AST) -> ast.AST:
        self._scope_level += 1
        try:
            return self.generic_visit(node)
        finally:
            self._scope_level -= 1

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        return self._enter_scope(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> ast.AST:
        return self._enter_scope(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.AST:
        return self._enter_scope(node)

    def visit_Lambda(self, node: ast.Lambda) -> ast.AST:
        return self._enter_scope(node)

    def visit_Return(self, node: ast.Return) -> ast.AST:
        if self._scope_level:
            return self.generic_visit(node)
        value = node.value
        if value is None:
            value = ast.Constant(value=None)
        else:
            value = self.visit(value)
        ret = ast.Raise(
            exc=ast.Call(
                func=ast.Name(id="ReturnValue", ctx=ast.Load()),
                args=[value],
                keywords=[],
            ),
            cause=None,
        )
        return ast.copy_location(ret, node)


def make_proc(code: str, path: Any, *, use_async: bool = False) -> CodeType:
    """
    Compile snippet code directly for eval/await execution.
    """
    flags = ast.PyCF_ALLOW_TOP_LEVEL_AWAIT if use_async else 0
    tree = compile(code, str(path), "exec", flags=flags | ast.PyCF_ONLY_AST)
    tree = _ReturnRewriter().visit(tree)
    ast.fix_missing_locations(tree)
    return compile(tree, str(path), "exec", flags=flags)


class Code:
    """
    Call wrapper for code snippets stored under ``code.exec``.

    The snippets are loaded from ``path`` and compiled lazily. This object
    transparently handles async code, direct sync code, and thread-offloaded
    sync code.
    """

    link: LinkSender
    path: Path

    _proc: CodeType | None = None
    _is_async: bool | None = None
    _vars: dict[str, Any] = {}
    _data: Mapping[str, Any] | Any = NotGiven

    def __init__(self, link: LinkSender, path: Path):
        self.link = link
        self.path = Path.build(path)

    def _load(self, data: Mapping[str, Any]) -> None:
        """Compile code data if no compiled process is cached."""
        if self._proc is not None:
            return
        if not isinstance(data, Mapping):
            raise TypeError(f"{self.path}: record must be a mapping")

        code = data["code"]
        is_async = data.get("is_async", None)
        self._proc = make_proc(code, self.path / "code", use_async=is_async is True)
        self._is_async = is_async

    def update(self, new_data: Any, compile: bool = False) -> None:  # noqa: A002
        """
        Replace stored record data and optionally compile it immediately.
        """
        self._data = new_data
        self._proc = None
        self._is_async = None
        if compile:
            self._load(new_data)

    def _globals(self, kw: dict[str, Any]) -> dict[str, Any]:
        """
        Build default globals exposed to snippet code.
        """
        return dict(
            runner=self,
            link=self.link,
            path=self.path,
            args=(),
            kw=kw,
            anyio=anyio,
            P=P,
            Path=Path,
            NotGiven=NotGiven,
            ReturnValue=ReturnValue,
        )

    @staticmethod
    def _exec(proc: CodeType, vars_: dict[str, Any]) -> Any:
        """Execute one compiled snippet with explicit locals."""
        return eval(proc, vars_)

    async def __call__(self, **kw: Any) -> Any:
        """
        Execute the snippet and return its result.
        """
        if self._proc is None:
            if self._data is NotGiven:
                self._data = await self.link.d_get(self.path)
            self._load(self._data)
        assert self._proc is not None  # set by _load

        call_kw = combine_dict(kw, self._data.get("vars", {}), self._globals(kw))

        if self._is_async is False:
            try:
                await run_sync(self._exec, self._proc, call_kw)
            except ReturnValue as exc:
                return exc.value
            return None

        try:
            res = self._exec(self._proc, call_kw)
            if self._is_async is True:
                await res
        except ReturnValue as exc:
            return exc.value
        return None



# command line interface  # noqa: D100
from __future__ import annotations

import anyio
from contextlib import nullcontext

import asyncclick as click

from moat.util import NotGiven, yprint
from moat.lib.path import P
from moat.lib.run import AliasedGroup, attr_args
from moat.link._data import data_get, node_attr
from moat.link.client import Link
from moat.link.meta import MsgMeta


@click.group(cls=AliasedGroup, short_help="Manage data.", invoke_without_command=True)  # pylint: disable=undefined-variable
@click.argument("path", type=P, nargs=1)
@click.pass_context
async def cli(ctx, path):
    """
    This subcommand reads and manipulates errors stored in the MoaT-Link service.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    obj.conn = await ctx.with_async_resource(Link(cfg))
    path = P("error") + path
    if ctx.invoked_subcommand is None:
        await data_get(obj.conn, path, meta=True, out=obj.stdout, recursive=False)
    else:
        obj.path = path


@cli.command
@click.pass_obj
async def get(obj, **k):
    """
    Retrieve a MoaT-Link error entry.

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """
    await data_get(obj.conn, obj.path, meta=obj.meta, **k)


@cli.command("set", short_help="Update an entry")
@click.option("-k/-K", "--ok/--bad", is_flag=True, help="Mark as ok / not ok")
@attr_args
@click.pass_obj
async def set_(obj, ok, **kw):
    """
    Update some MoaT-Link error entry.
    """
    if ok is not None:
        kw["eval_"]["ok"] = repr(ok)
    res = await node_attr(obj, obj.path, **kw)

    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command(short_help="Delete an entry / subtree")
@click.option(
    "-b",
    "--before",
    type=float,
    help="Don't delete entries created after this timestamp",
)
@click.option("-r", "--recursive", is_flag=True, help="Delete a complete subtree")
@click.pass_obj
async def delete(obj, before, recursive):
    """
    Delete an entry, or a subtree.

    You might want to use "--before" flag to ensure that no other change
    arrived after you viewed the data in question.

    Non-recursively deleting an entry with children works and does *not*
    affect the child entries.

    The root entry cannot be deleted.
    """
    args = {}
    if recursive:
        args["rec"] = recursive
    if before:
        args["ts"] = before
    res = await obj.conn.d.delete(obj.path, **args)
    if obj.meta:
        res = dict(data=res[0], meta=MsgMeta.restore(res[1:]).repr())
    else:
        res = res[0]
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-m", "--mode", type=str, help="Retrieval mode", default="s")
@click.option("-M", "--mark", is_flag=True, help="Retrieval mode")
@click.option("-s", "--subtree", is_flag=True, help="Read the whole tree.")
@click.option("-D", "--add-date", is_flag=True, help="Add *_date entries")
@click.option("-i", "--ignore", multiple=True, type=P, help="Skip this (sub)tree")
@click.option("-n", "--min-length", type=int, help="Minimum path length")
@click.option("-N", "--max-length", type=int, help="Maximum path length")
@click.option("-a", "--max-age", type=int, help="Skip entries older than N seconds")
@click.option("-t", "--timeout", type=int, help="Stop reading after N seconds")
@click.pass_obj
async def monitor(
    obj,
    mode,
    add_date,  # noqa: ARG001
    ignore,
    mark,
    subtree,
    min_length,
    max_length,
    max_age,
    timeout,
):
    """Monitor a MoaT-Link subtree.

    The mode can be:
    * c/current   read current data from the server
    * u/update    read updates from MQTT
    * s/stream    current plus updates
    * m/mqtt      subscribe to MQTT stream, including retained data
    """

    match mode:
        case "c" | "current":
            state = True
        case "u" | "update":
            state = False
        case "s" | "stream":
            state = None
        case "m" | "mqtt":
            state = NotGiven
        case _:
            raise click.UsageError("Mode must be current|update|stream|mqtt")
    if mark and state is not None:
        raise click.UsageError("You can only add a mark in Stream mode")

    def pm(p):
        for ip in ignore:
            if len(p) >= len(ip) and p[: len(ip)] == ip:
                return True
        return False

    with anyio.move_on_after(timeout) if timeout else nullcontext():
        async with obj.conn.d_watch(
            obj.path,
            state=state,
            mark=mark,
            meta=True,
            subtree=subtree,
            max_length=max_length,
            min_length=min_length,
            age=max_age,
        ) as mon:
            async for res in mon:
                if res is None:
                    res = "*** Snapshot data ends ***"  # noqa:PLW2901
                yprint(res, stream=obj.stdout)
                print("---", file=obj.stdout)
                obj.stdout.flush()



from __future__ import annotations  # noqa: D100

import anyio
import logging
import time
from contextlib import asynccontextmanager, suppress
from enum import Enum, auto
from functools import partial

from attrs import define, field

from moat.util import CtxObj, NotGiven, attrdict, srepr
from moat.lib.broadcast import Broadcaster
from moat.lib.path import P
from moat.lib.priomap import TimerMap

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.path import Path
    from moat.link.client import Link

    from collections.abc import Awaitable
    from typing import Any

__all__ = ["HostList", "ServiceMon"]

logger = logging.getLogger(__name__)


#
# Monitoring
#
class HostState(Enum):
    # constants
    INIT = "in"
    NEW = "nw"
    DOWN = "dn"  # id state=down seen
    UP = "up"
    TIMEOUT = "tm"  # no ping for some time, error generated
    STALE = "sl"  # superseded
    DROP = "xx"  # no ping for a long time, entry dropped

    # We want "id" and "ping" messages.
    # "host" is an additional "name this host" signal.
    ONLY_I = "oi"  # only host/id seen
    ONLY_P = "op"  # only ping seen


class HostEvent(Enum):
    INIT = auto()
    MSG_HOST = auto()
    MSG_PING = auto()
    MSG_ID = auto()

    MSG_DOWN = auto()
    DEL_HOST = auto()  # Service gets dropped, or replaced
    DEL_ID = auto()

    TIMEOUT = auto()


_S = HostState
_E = HostEvent


@define
class _HostTransition:
    source: HostState
    dest: HostState | None


@define
class _HostEventData:
    model: HostMachine
    event: HostEvent
    transition: _HostTransition
    kwargs: dict[str, Any] = field(factory=dict)


class _ServiceData(attrdict):
    h: dict[Path, dict[str, Any]]
    i: Any
    p: dict[str, Any]


class HostMachine:
    def __init__(self, host):
        self.host = host
        self.state = HostState.INIT

    def _base_state(self) -> HostState:
        data = self.host.data
        has_id = "i" in data
        has_ping = "p" in data
        has_host = bool(data.h)
        has_identity = has_id or has_host

        if not has_identity and not has_ping:
            return HostState.DROP
        if not has_ping:
            return HostState.ONLY_I
        if not has_identity:
            return HostState.ONLY_P
        if not data.p.get("up", False):
            return HostState.DOWN if has_id else HostState.ONLY_P
        return HostState.UP

    async def _emit(self, event: HostEvent, dest: HostState | None, **kwargs) -> None:
        evt = _HostEventData(
            model=self,
            event=event,
            transition=_HostTransition(source=self.state, dest=dest),
            kwargs=kwargs,
        )
        self.host.mon.set_timeout(self.host)
        self.host.updated(evt)

    async def _set_state(self, new_state: HostState, event: HostEvent, **kwargs) -> None:
        if new_state is self.state:
            await self._emit(event, None, **kwargs)
            return

        old_state = self.state
        evt = _HostEventData(
            model=self,
            event=event,
            transition=_HostTransition(source=old_state, dest=new_state),
            kwargs=kwargs,
        )
        if (cb := getattr(self.host, f"on_exit_{old_state.name.lower()}", None)) is not None:
            await cb(evt)
        self.state = new_state
        if (cb := getattr(self.host, f"on_enter_{new_state.name.lower()}", None)) is not None:
            await cb(evt)
        self.host.mon.set_timeout(self.host)
        self.host.updated(evt)
        if new_state is HostState.DROP:
            await self.host.drop_both(evt)

    async def trigger(self, event: HostEvent, **kwargs) -> None:
        if self.state is HostState.INIT:
            if event is _E.INIT:
                await self._set_state(_S.NEW, event, **kwargs)
            return

        if self.state is HostState.DROP:
            if event in (_E.MSG_HOST, _E.MSG_ID, _E.MSG_PING, _E.MSG_DOWN, _E.DEL_HOST):
                evt = _HostEventData(
                    model=self,
                    event=event,
                    transition=_HostTransition(source=self.state, dest=None),
                    kwargs=kwargs,
                )
                await self.host.re_init(evt)
            elif event is _E.DEL_ID:
                await self.host.drop_id(None)
            elif event is _E.TIMEOUT:
                await self.host.drop_both()
            return

        if event is _E.TIMEOUT:
            if self.state in (_S.NEW, _S.ONLY_P):
                dest = _S.DROP
            elif self.state in (_S.ONLY_I, _S.UP, _S.DOWN):
                dest = _S.DROP if not self.host.data.h else _S.TIMEOUT
            elif self.state is _S.TIMEOUT:
                dest = _S.DROP if not self.host.data.h else _S.STALE
            elif self.state is _S.STALE:
                dest = _S.DROP
            else:
                dest = self.state
            await self._set_state(dest, event, **kwargs)
            return

        if event in (_E.DEL_ID, _E.DEL_HOST) and self.state in (_S.TIMEOUT, _S.STALE):
            await self._emit(event, None, **kwargs)
            return

        await self._set_state(self._base_state(), event, **kwargs)


@define(eq=False)
class Service:
    """Contains data for one particular service."""

    mon: HostList = field()
    id: str = field()

    # These get filled via path
    data: _ServiceData = field(factory=_ServiceData, init=False)
    machine: HostMachine = field(init=False)
    _lock: anyio.Lock = field(init=False, factory=anyio.Lock, repr=False)

    _last = field(default=0, init=False)

    def __attrs_post_init__(self):
        self.machine = HostMachine(self)
        self.data.h = dict()

    @property
    def state(self):
        return self.machine.state

    @property
    def timeout(self):
        "returns the timeout under `cfg.link.timout.ping`"
        if self.state is HostState.DROP:
            return 0  # error when looked up

        pto = self.mon.cfg.timeout.ping
        if self.state in (HostState.NEW, HostState.ONLY_I, HostState.ONLY_P):
            return pto.stale
        if self.state in (HostState.UP, HostState.DOWN):
            return pto.timeout
        if self.state is HostState.TIMEOUT:
            if not self.last:
                return pto.delete
            age = time.time() - self.last
            return max(pto.delete, 0.0, pto.delay - age)
        if self.state is HostState.STALE:
            return pto.delay
        return pto.timeout

    def updated(self, evt):
        "queue update to Broadcaster"
        self.mon.updated(self, evt)

    def drop_id(self, _evt) -> Awaitable[None]:
        return self.mon.drop_id(self)

    def drop_host(self, _evt) -> Awaitable[None]:
        return self.mon.drop_host(self)

    async def drop_both(self, evt=None) -> None:
        evt  # noqa:B018
        await self.mon.drop_host(self)
        await self.mon.drop_id(self)

    @property
    def last(self):
        return self._last

    @last.setter
    def last(self, tm):
        if self._last < tm:
            self._last = tm

    async def trigger(self, *args, **kwargs) -> None:
        async with self._lock:
            await self.machine.trigger(*args, **kwargs)

    async def on_enter_drop(self, ev):  # noqa: ARG002
        await self.mon.drop_cb(self)

    async def on_enter_timeout(self, ev):  # noqa: ARG002
        await self.mon.drop_id(self)

    async def on_enter_stale(self, ev):  # noqa: ARG002
        await self.mon.drop_host(self)

    async def re_init(self, ev):
        """
        We have an incoming packet while a DROP event is running.
        This should not happen, but whatever …
        """
        ho = ev.model.host
        mon = ho.mon
        # dprint("**RE*",ho.id)
        h = Service(mon, ho.id)
        h.data = ho.data
        # not copying the paths!
        mon.ids[h.id] = h
        await h.trigger(_E.INIT)


class HostList(CtxObj):
    """
    This class monitors the 'run.ping', 'run.id' and 'run.host' channels.
    """

    def __init__(
        self,
        cfg: dict,
        link: Link,
        debug: bool = False,
        broadcaster: Broadcaster[Any] | None = None,
    ):
        self.link = link
        self.debug = debug
        self.cfg = cfg
        self._bc = broadcaster

        # Service Instance
        self.hsi: dict[Path, Service] = {}
        self.ids: dict[str, Service] = {}
        self.times: TimerMap[str] = TimerMap()

    @asynccontextmanager
    async def _ctx(self):
        if self._bc is None:
            self._bc = Broadcaster(10)

        async with self._bc, anyio.create_task_group() as self.tg:
            await self.start_tasks()
            yield self._bc
            self.tg.cancel_scope.cancel()

    async def start_tasks(self):
        "Starting Tasks. Can be supplanted by subclasses."
        self.tg.start_soon(self._mon_host)
        self.tg.start_soon(self._mon_ping)
        self.tg.start_soon(self._mon_id)
        self.tg.start_soon(self._timer)

    def set_timeout(self, host: Service, to: float | None = None):
        "Update service entry in timeout queue"
        if to is None:
            to = host.timeout
        if not to:
            with suppress(KeyError):
                self.times.pop(host.id)
            return

        # dprint("TIME ", host.id, to)
        self.times[host.id] = to

    async def _timer(self):
        async for id in self.times:
            try:
                h = self.ids[id]
            except KeyError:
                continue
            self.tg.start_soon(h.trigger, _E.TIMEOUT)

    def updated(self, host: Service, evt):
        "queue update to Broadcaster"
        if evt.transition.dest is not None or evt.kwargs.get("changed", False):
            logger.debug("Service %s: %s %s", host.id, host.state.name, srepr(host.data))
            self._bc(host)

    async def _mon_host(self):
        """
        Monitor host messages.

        These update the .hsi lookup but don't actually affect state.
        """
        async with self.link.d_watch(P("run.host"), subtree=True, state=NotGiven) as mon:
            async for p, msg in mon:
                try:
                    if msg is NotGiven:  # host deleted
                        with suppress(KeyError, AttributeError):
                            h = self.hsi.pop(p)
                            await self.drop_path(h, p)
                            if not h.data.h:
                                self.tg.start_soon(h.trigger, _E.DEL_HOST)

                        continue

                    id = msg.pop("id")

                    # points to another service? drop it from the other entry
                    with suppress(KeyError, AttributeError):
                        h = self.hsi[p]
                        if h.id != id:
                            await self.drop_path(h, p)
                            if not h.data.h:
                                self.tg.start_soon(h.trigger, _E.DEL_HOST)
                            # unconditionally updated below

                    try:
                        h = self.ids[id]
                    except KeyError:
                        self.ids[id] = h = Service(mon=self, id=id)
                        await h.trigger(_E.INIT)
                    self.hsi[p] = h
                    changed = h.data.h.get(p, None) != msg
                    self.add_path(h, p, msg)
                    self.tg.start_soon(partial(h.trigger, _E.MSG_HOST, changed=changed))
                    # dprint("H    ",p,id,h)

                except Exception as exc:
                    logger.error("BadHost %r %r", p, msg, exc_info=exc)

    async def _mon_id(self):
        async with self.link.d_watch(P("run.id"), subtree=True, meta=True, state=NotGiven) as mon:
            async for p, msg, meta in mon:
                try:
                    if len(p) != 1:
                        logger.warning("BadID %r %r", p, msg)
                        continue

                    id = p[0]
                    if msg is NotGiven:  # ID deleted
                        with suppress(KeyError):
                            h = self.ids[id]
                            changed = "i" in h.data
                            h.data.pop("i", None)
                            self.tg.start_soon(partial(h.trigger, _E.DEL_ID, changed=changed))
                        continue

                    try:
                        h = self.ids[id]
                    except KeyError:
                        self.ids[id] = h = Service(mon=self, id=id)
                        await h.trigger(_E.INIT)
                    h.last = meta.timestamp

                    changed = False
                    if h.data.get("i", None) != msg:
                        h.data.i = msg
                        changed = True
                    self.tg.start_soon(partial(h.trigger, _E.MSG_ID, changed=changed))
                    # dprint("P    ",p,msg)

                except Exception as exc:
                    logger.error("BadId %r %r", p, msg, exc_info=exc)

    async def _mon_ping(self):
        async with self.link.d_watch(P("run.ping.id"), subtree=True, state=NotGiven) as mon:
            async for p, msg in mon:
                try:
                    if len(p) != 1 or not isinstance(msg, dict):
                        logger.warning("BadPING %r %r", p, msg)
                        continue
                    id = p[0]

                    try:
                        h = self.ids[id]
                    except KeyError:
                        self.ids[id] = h = Service(mon=self, id=id)
                        await h.trigger(_E.INIT)

                    changed = False
                    if h.data.get("p", None) != msg:
                        h.data.p = msg
                        changed = True
                    self.tg.start_soon(
                        partial(
                            h.trigger, _E.MSG_PING if msg["up"] else _E.MSG_DOWN, changed=changed
                        )
                    )
                    # dprint("P    ",p,msg)

                except Exception as exc:
                    logger.error("BadPing %r %r", p, msg, exc_info=exc)

    async def drop_cb(self, host):
        """
        Called from `Service.on_enter_drop`
        """
        # dprint("  DCB",host.id)
        for p in host.data.h:
            if self.hsi.get(p, None) is host:
                del self.hsi[p]
        if self.ids.get(host.id, None) is host:
            del self.ids[host.id]
        self.set_timeout(host, False)

    async def drop_id(self, host):
        """
        Seen a message that deletes this ID entry.
        """
        host.data.pop("i", None)

    def add_path(self, host: Service, path: Path, msg: dict):
        """
        Add a path with this message to the service
        """
        host.data.h[path] = msg

    async def drop_host(self, host: Service):
        """
        Seen a message that deletes this service entry.
        """
        for p in tuple(host.data.h):
            if self.hsi.get(p, None) is host:
                del self.hsi[p]
        host.data.h.clear()

    async def drop_path(self, host: Service, path: Path):
        """
        Seen a message that deletes/supersedes this service entry.
        """
        host.data.h.pop(path, None)


class ServiceMon(HostList):
    """
    The Service Monitor runs once in a MoaT-Link network, as part of the main
    hosts's 'moat-link-host' service.

    Its main job is to remove stale retained entries under 'run.id' and 'run.host'.
    """

    def __init__(self, *a, fake: bool = False, **kw):
        """ """
        super().__init__(*a, **kw)
        self.hostdown: TimerMap[Path] = TimerMap()
        self.hostup: TimerMap[Path] = TimerMap()
        self.errored: dict[Path, bool] = dict()
        self.fake: bool = fake

    async def start_tasks(self):
        "internal helper"
        await super().start_tasks()
        self.tg.start_soon(self._mon_hostdown)
        self.tg.start_soon(self._mon_hostup)

    async def _mon_hostdown(self):
        # watch host comings+goings, complain if one is down for too long
        # (TODO: or goes down+up too often)
        async for path in self.hostdown:
            # print("******** TIME",path)
            if path in self.hsi:
                # Present. Clear error.
                await self._no_err(path)
            else:
                # not present. Error.
                await self._err(None, path, "down")

    async def _mon_hostup(self):
        # watch host.data.up
        # (TODO: or goes down+up too often)
        async for path in self.hostup:
            # print("******** TIMEUP",path)
            try:
                host = self.hsi[path]
            except KeyError:
                # not present.
                pass
            else:
                # Present. Clear error.
                if not host.data.h[path].get("up", False):
                    await self._err(None, path, "not up")

    async def drop_id(self, host):
        "Delete a host's ID message"
        if "i" in host.data:
            if self.fake:
                print("DEL", host.id, host)
            else:
                await self.link.d_set(P("run.id") / host.id, retain=True)
        await super().drop_id(host)

    async def drop_host(self, host):
        "Delete a host's Service messages (yes all of them)"
        for p in host.data.h.keys():
            if self.fake:
                print("DEL", p)
            else:
                await self.link.d_set(P("run.host") + p, retain=True)
            self.hostdown[p] = self.cfg.timeout.restart.error
            with suppress(KeyError):
                del self.hostup[p]
        await super().drop_host(host)

    async def _err(self, host: Service | None, path: Path, msg: str, data: Any = None):
        if self.errored.get(path, "") != msg:
            dat = {"msg": msg, "level": 4}
            if data is not None:
                dat["aux"] = data
            if host is not None and (dt := host.data.h.get(path, None)) is not None:
                dat["data"] = dt
            if self.fake:
                print("ERR", path, dat)
            else:
                await self.link.d_set(P("error.run.host") + path, dat)
            self.errored[path] = msg

    async def _no_err(self, path: Path):
        if self.errored.get(path, True) is not False:
            if self.fake:
                print("OK ", path)
            else:
                await self.link.d_set(P("error.run.host") + path)
            self.errored[path] = False

    def add_path(self, host: Service, path: Path, msg: dict):
        "Add a path with this message to the service"
        # print("******** ADD",path)
        super().add_path(host, path, msg)
        self.hostdown[path] = self.cfg.timeout.restart.flap
        if msg.get("up", False):
            with suppress(KeyError):
                del self.hostup[path]
        else:
            self.hostup[path] = self.cfg.timeout.restart.up

    async def drop_path(self, host: Service, path: Path):
        "Remove a path with this message from the service"
        # print("******** DROP",path)
        try:
            msg = host.data.h[path]
        except KeyError:
            msg = None
        try:
            self.hostdown.pop(path)
        except KeyError:
            self.hostdown[path] = self.cfg.timeout.restart.error
        else:
            if msg is not None and msg.get("up", False):
                await self._err(host, path, "flapping", msg)
            else:
                self.hostdown[path] = self.cfg.timeout.restart.error
        await super().drop_path(host, path)



"""
Bare-bones connection to a MoaT server
"""

from __future__ import annotations

import anyio
import logging

from attrs import define, field

from moat.util import attrdict
from moat.lib.path import P

from . import protocol_version as proto_version
from . import protocol_version_min as proto_version_min
from .common import CmdCommon

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import Msg, MsgSender

    from .auth import AuthMethod

    from typing import Any

logger = logging.getLogger(__name__)


class NotAuthorized(RuntimeError):  # noqa: D101
    pass


def _to_dict(x: dict[str, AuthMethod] | list[AuthMethod]) -> dict[str, AuthMethod]:
    if isinstance(x, dict):
        return x
    return {a.name: a for a in x}


class _NoRPCAuth(RuntimeError):
    "Internal marker for falling back to legacy Hello/auth."


class _RPCRoot:
    "Minimal object exposing ``.sender`` for the RPC auth helper."

    def __init__(self, sender):
        self.sender = sender


class _RPCShim:
    "Adapter exposing the subset of BaseCmdMsg used by moat.lib.rpc.auth."

    def __init__(self, hello: Hello, sender, auth_data: dict[str, Any]):
        self.hello = hello
        self._sender = sender
        self.auth = attrdict(auth_data)
        self.is_server = (
            hello.rpc_auth_server if hello.rpc_auth_server is not None else hello.me_server
        )

    @property
    def auth_name(self):
        "Name presented to the RPC auth helper."
        return self.hello.me

    async def wait_ready(self, wait: bool = True):
        "The live RPC stream is already running."
        wait  # noqa:B018
        return False

    async def handle(self, msg, rcmd, _auth: bool = False):
        "Forward auth helper traffic to the already-open RPC sender."
        _auth  # noqa:B018
        return await self._sender.handle(msg, rcmd)

    def auth_data_out(self) -> dict:
        "Delegate extra auth-request data handling to Hello."
        return self.hello.auth_data_out()

    def auth_data_in(self, args, data: dict) -> None:
        "Delegate extra incoming auth-request data handling to Hello."
        self.hello.auth_data_in(args, data)

    def auth_data_res_out(self, role: str) -> dict:
        "Delegate extra auth-response data handling to Hello."
        return self.hello.auth_data_res_out(role)

    def auth_data_res_in(self, role: str, data: dict) -> None:
        "Delegate extra incoming auth-response data handling to Hello."
        self.hello.auth_data_res_in(role, data)

    def auth_skip(self) -> None:
        "Signal that the peer does not support RPC auth."
        self.hello.auth_skip()


@define
class Hello(CmdCommon):
    """
    This object handles the initial handshake between two MoaT links.

    Usage:

    * The ``handle`` method that you supplied to your `MsgHandler` must
      forward all incoming commands to `Hello.handler` while
      ``auth_data`' is `None`.

    * Call `Hello.run`.

    Note that due to the async nature of the protocol, additional commands
    may arrive even before `Hello.run` has returned.

    Negotiated auth data are in ``.auth_data``.
    """

    if TYPE_CHECKING:

        def __init__(
            self,
            me: str | None = None,
            them: str | None = None,
            *,
            auth_in: dict[str, AuthMethod] | list[AuthMethod] = ...,
            auth_out: dict[str, AuthMethod] | list[AuthMethod] = ...,
            rpc_auth_modes: tuple[str, ...] | None = None,
            rpc_auth_data: dict[str, Any] = ...,
            rpc_auth_server: bool | None = None,
            me_server: bool = False,
            protocol_min: int = proto_version_min,
            protocol_max: int = proto_version,
        ) -> None: ...

    me: str | None = field(default=None)
    them: str | None = field(default=None)

    auth_data: Any = field(init=False, default=None)

    auth_in: dict[str, AuthMethod] = field(kw_only=True, default={}, converter=_to_dict)  # noqa:RUF008
    auth_out: dict[str, AuthMethod] = field(kw_only=True, default={}, converter=_to_dict)  # noqa:RUF008
    rpc_auth_modes: tuple[str, ...] | None = field(kw_only=True, default=None)
    rpc_auth_data: dict[str, Any] = field(kw_only=True, factory=dict)
    rpc_auth_server: bool | None = field(kw_only=True, default=None)

    me_server: bool = field(default=False)
    they_server: bool = field(init=False, default=False)

    _sync: anyio.Event = field(init=False, factory=anyio.Event)
    _done: anyio.Event = field(init=False, factory=anyio.Event)

    # min and max protocol versions we might accept
    protocol_min: int = field(kw_only=True, default=proto_version_min)
    protocol_max: int = field(kw_only=True, default=proto_version)

    # negotiated protocol version
    protocol_version: int = field(init=False, default=0)
    _rpc_auth: Any = field(init=False, default=None)
    _rpc_in: Any = field(init=False, default=None)
    _rpc_fallback: bool = field(init=False, default=False)
    _name_reply: str | None = field(init=False, default=None)
    _res_data: dict[str, Any] | None = field(init=False, default=None)

    def __attrs_post_init__(self):
        if self.me_server and self.me is None:
            raise ValueError("A server must have a name")

    def auth_data_out(self) -> dict:
        """
        Extra auth-request metadata shared between RPC auth and legacy Hello.
        """
        res = {"version": proto_version}
        if self.them:
            res["rname"] = self.them
        return {"moat.link": res}

    def auth_data_in(self, args, data: dict) -> None:
        """
        Process incoming auth-request metadata.

        Args:
            args: auth/hello positional arguments, with the peer name at index 2.
            data: auth/hello keyword metadata.
        """
        info = data.get("moat.link", {})
        if info is None:
            info = {}
        try:
            prot = info.get("version", args[0])
        except IndexError:
            prot = info.get("version", None)
        if prot is not None:
            if prot < self.protocol_min:
                raise ValueError("Protocol mismatch", prot)
            self.protocol_version = min(prot, self.protocol_max)

        try:
            self.they_server = args[1]
        except IndexError:
            pass

        local_name = info.get("rname", None)
        remote_name = args[2] if len(args) > 2 else None
        if remote_name is None:
            if self.them is None:
                logger.error("No remote name")
            else:
                self._name_reply = self.them
        elif self.them is None:
            self.them = remote_name
        elif self.them != remote_name:
            logger.debug("Remote name: %r / %r", remote_name, self.them)
            self._name_reply = self.them
            self.them = remote_name

        if local_name is None:
            return
        if self.me is None:
            self.me = local_name
        elif self.me != local_name:
            logger.debug("My name: %r / %r", local_name, self.me)

    def auth_data_res_out(self, role: str | None = None) -> dict:
        """
        Extra auth-response metadata shared between RPC auth and legacy Hello.
        """
        role  # noqa:B018
        if self._name_reply is None:
            return {}
        return {"name": self._name_reply}

    def auth_data_res_in(self, role: str | None, data: dict) -> None:
        """
        Process incoming auth-response metadata.
        """
        role  # noqa:B018
        data = dict(data)
        if (name := data.pop("name", None)) is not None:
            self.me = name
        self._res_data = data or None

    def auth_skip(self) -> None:
        """
        RPC auth isn't supported by the peer; caller may fall back to Hello/auth.
        """
        raise _NoRPCAuth

    def _rpc_modes(self) -> tuple[str, ...]:
        """
        RPC auth methods to try before the legacy hello/auth handshake.
        """
        if self.rpc_auth_modes is None:
            return ()
        return tuple(self.rpc_auth_modes)

    async def _run_rpc(self, sender: MsgSender):
        """
        Try the RPC auth protocol and return Hello-compatible result data.
        """
        modes = self._rpc_modes()
        if not modes:
            raise _NoRPCAuth

        from moat.lib.rpc.auth._base import Auth, AuthCmdIn, AuthDenied  # noqa: PLC0415
        from moat.lib.rpc.errors import RemoteError  # noqa: PLC0415

        cfg = attrdict(modes=[attrdict(mode=m) for m in modes])
        shim = _RPCShim(self, sender, self.rpc_auth_data)
        auth = Auth(cfg, cast("Any", shim))
        auth.base_root = cast("MsgSender", _RPCRoot(sender))
        a_in = AuthCmdIn(auth)
        self._rpc_auth = auth
        self._rpc_in = a_in
        self._rpc_fallback = False
        self._name_reply = None
        self._res_data = None
        try:
            async with a_in:
                await a_in.task()
            if self._rpc_fallback:
                raise _NoRPCAuth
            if isinstance(auth.ok, Exception):
                raise auth.ok
            self.auth_data = getattr(auth.ok, "name", True)
            if a_in.p_version is not None:
                self.protocol_version = min(a_in.p_version, self.protocol_max)
            self._done.set()
            return self._res_data or True
        except AuthDenied:
            self.auth_data = False
            self._done.set()
            raise
        except _NoRPCAuth:
            raise
        except BaseException as exc:
            if isinstance(exc, RemoteError) and exc.args == ("ValueError", "No Hello/Auth"):
                raise _NoRPCAuth from None
            if isinstance(exc, ValueError) and exc.args == ("No Hello/Auth",):
                raise _NoRPCAuth from None
            if self._rpc_fallback:
                raise _NoRPCAuth from None
            self.auth_data = False
            self._done.set()
            raise
        finally:
            self._rpc_in = None
            self._rpc_auth = None

    async def handle(self, msg: Msg, rcmd: list[PathElem]) -> None:
        """
        Dispatch an incoming "hello" message
        """
        if self._rpc_in is not None and rcmd and rcmd[-1] is None:
            # new auth protocol
            return await self._rpc_in.handle(msg, rcmd)
        scmd = rcmd.pop()
        if scmd != "i":
            raise KeyError(scmd)
        if len(rcmd) == 1 and rcmd[0] == "hello":
            if self._rpc_in is not None:
                # The peer is using the legacy startup protocol. Let the incoming
                # hello proceed immediately, then fall back once our pending RPC
                # auth attempt receives its rejection.
                self._rpc_fallback = True
                self._sync.set()
            res = await self.do_hello(msg)
            return await msg.result(res)
        if len(rcmd) != 2 or rcmd[1] != "auth":
            raise ValueError("No Hello/Auth")

        if self.auth_data is not None:
            # Some other method already succeeded
            await msg.result(False)
            return
        rcmd0 = rcmd[0]
        if not isinstance(rcmd0, str):
            await msg.result(False)
            return
        a = self.auth_in.get(rcmd0, None)
        if a is None:
            await msg.result(False)
        else:
            await a.handle(self, msg)

    def authorized(self, data: Any) -> bool:
        """
        Called by an auth method to indicate that authorization worked.

        Returns True if this method was the first to succeed.
        """
        if self.auth_data is not None:
            return False
        self.auth_data = data
        return True

    doc_i_hello = dict(
        _d="Process remote Hello msg",
        _r="auth state",
        _0="int:protocol",
        _1="bool:server flag",
        _2="str:sender's name",
        _3="str:recipient's name (temp by sender)",
        _4=["str:auth method"],
        _k="str:auth names",
        _kw="Any:auth params",
    )

    async def do_hello(self, msg) -> bool | None:
        """
        Process the remote hello message.

        Returns True if no auth is required.
        """
        try:
            res = await self._do_hello(msg)
        except BaseException:
            self.auth_data = False
            raise
        else:
            if self.auth_data is None:
                self.auth_data = res
            await msg.result(res)
        finally:
            self._done.set()

    async def _do_hello(self, msg) -> bool | dict:
        logger.debug("H IN %r %r", msg.args, msg.kw)
        it = iter(msg.args)
        auth = True
        remote_name = None
        local_name = None

        try:
            prot = next(it)

            # TODO special auth for servers?
            they_server = next(it)
            self.they_server = they_server
            if not self.they_server and not self.me_server:
                raise RuntimeError("Two clients cannot talk")

            remote_name = next(it)
            self._name_reply = None
            try:
                local_name = next(it)
            except StopIteration:
                self.auth_data_in((prot, they_server, remote_name), {})
                raise
            else:
                res = {}
                if local_name is not None:
                    res["rname"] = local_name
                self.auth_data_in(
                    (prot, they_server, remote_name),
                    {"moat.link": res},
                )
            if remote_name is None and self.them is None:
                auth = False
                raise StopIteration

            auth = next(it)

            if not next(it):
                raise RuntimeError("Not talking to a server")

        except StopIteration:
            pass

        # wait for the outgoing part to start
        await self._sync.wait()

        if auth is False:
            raise NotAuthorized("Remote blocks us", self.them)
        if auth is True:
            self.auth_data = True
            aux_data = self.auth_data_res_out(None)
            return aux_data or True

        if isinstance(auth, str):
            auth = (auth,)

        # Check for auth data in the Hello
        aux_data = {}
        for a in self.auth_in.values():
            res = await a.hello_in(self, msg.kw.get(a.name, None))
            if res is False:
                return False
            if res:
                if self.auth_data is None:
                    self.auth_data = True
                aux_data.update(self.auth_data_res_out(a.name))
                break

        # cycle through the remote side's accepted auth methods
        for a in auth:
            am = self.auth_out.get(a, None)
            if am is None:
                continue
            res = await am.chat(self, msg.kw.get(a, None))
            if res is None:
                continue
            aux_data = self.auth_data_res_out(am.name)
            if isinstance(res, dict):
                aux_data.update(res)
            if res is False or not aux_data:
                return res
            return aux_data

        # Nothing matched.
        return False

    async def run(self, sender: MsgSender, **kw):
        """
        Send our Hello message.
        """
        try:
            return await self._run_rpc(sender)
        except _NoRPCAuth:
            pass

        auths = []
        for a in self.auth_in.values():
            auths.append(a.name)
            if a.name not in kw:
                v = await a.hello_out()
                if v is not None:
                    kw[a.name] = v
        kw.update(self.auth_data_out())

        if len(auths) == 0:
            auths = True
        elif len(auths) == 1:
            auths = auths[0]

        logger.debug("H OUT %d %s %s %r %r", proto_version, self.me, self.them, auths, kw)
        self._sync.set()
        res = await sender.cmd(
            P("i.hello"),
            proto_version,
            self.me_server,
            self.me,
            self.them,
            auths,
            **kw,
        )
        res = res.kw or res[0]

        if res is False:
            raise NotAuthorized("Server %r rejects us", self.them)
        if isinstance(res, dict):
            self.auth_data_res_in(None, res)

        # Wait for the incoming side of the auth/hello dance to succeed
        await self._done.wait()
        return res

    async def wait_done(self):
        "wait until done"
        await self._done.wait()



"""
Command-line interface for managing stored code snippets.
"""

from __future__ import annotations

import os

import asyncclick as click

from moat.util import NotGiven, edit_text, edit_yaml, yload, yprint
from moat.lib.path import P, Path, PathLongener
from moat.lib.run import AliasedGroup, attr_args, process_args
from moat.link.client import Link
from moat.link.code import CODE_EXEC_ROOT
from moat.link.code.run import make_proc
from moat.link.meta import MsgMeta

from collections.abc import Mapping
from typing import Any

EDIT_WHOLE = "w"
EDIT_CODE = "c"
EDIT_NON_CODE = "n"
EDIT_SAVE = "s"
EDIT_ABORT = "a"


def _sanitize_vars(value: Any) -> dict[str | None, Any]:
    "Normalize the configured argument defaults."
    if value in (NotGiven, None):
        return {}
    if not isinstance(value, Mapping):
        raise TypeError("vars must be a mapping")
    res: dict[str | None, Any] = {}
    for key, val in value.items():
        if key is None or isinstance(key, str):
            res[key] = val
        else:
            raise TypeError("vars keys must be strings")
    return res


def _check_exec_syntax(data: Mapping[str, Any], path: Path) -> None:
    """Compile the configured ``code`` script."""
    code = data.get("code", NotGiven)
    if code is NotGiven:
        raise KeyError(path / "code")
    if not isinstance(code, str):
        raise TypeError("code must be a string")
    _sanitize_vars(data.get("vars", {}))
    is_async = data.get("is_async", None)
    if is_async not in (None, True, False):
        raise TypeError("is_async must be true, false, or null")
    make_proc(code, path / "code", use_async=is_async is True)


def _info_line(data: Mapping[str, Any]) -> str:
    """Return an info line for a code entry."""
    info = data.get("info", NotGiven)
    if info is not NotGiven:
        return str(info)
    code = data.get("code", "")
    if isinstance(code, str):
        return f"<{len(code.splitlines())} lines>"
    return "<code>"


async def _list_entries(obj, as_dict, maxdepth, mindepth, full, short):
    """List code entries in a subtree."""
    if (full or as_dict) and short:
        raise click.UsageError("'-f'/'-d' and '-s' are incompatible.")

    args = [None, None, None]
    if mindepth is not None:
        args[1] = mindepth
    if maxdepth is not None:
        args[2] = maxdepth
    while args and args[-1] is None:
        args.pop()

    pl = PathLongener(obj.path)
    out: dict[Any, Any] | None = {} if as_dict is not None else None

    async with obj.conn.d.walk(obj.path, *args).stream_in() as mon:
        async for n, p, data, *_m in mon:
            path = pl.long(n, p)
            if not isinstance(data, Mapping):
                continue
            if "code" not in data:
                continue

            entry = dict(data)
            if not full:
                entry.pop("code", None)

            if short:
                print(path, "::", _info_line(data), file=obj.stdout)
                continue

            if as_dict is not None:
                assert out is not None
                cur: dict[Any, Any] = out
                for pp in path:
                    cur = cur.setdefault(pp, {})
                cur[as_dict] = entry
            else:
                yprint([{path: entry}], stream=obj.stdout)

    if as_dict is not None:
        yprint(out, stream=obj.stdout)


@click.group(cls=AliasedGroup, short_help="Manage code snippets.", invoke_without_command=True)
@click.option("-m", "--meta", is_flag=True, help="Include metadata")
@click.argument("path", type=P, nargs=1)
@click.pass_context
async def cli(ctx, path, meta):
    """
    Manage code snippets stored in MoaT-Link.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    if obj.get("port", None) is not None:
        cfg.client.port = obj.port
    obj.conn = await ctx.with_async_resource(Link(cfg, common=True))
    obj.meta = meta
    obj.path = CODE_EXEC_ROOT + path

    if ctx.invoked_subcommand is None:
        await _list_entries(obj, None, None, None, False, True)


@cli.command()
@click.option("-S", "--script", type=click.File(mode="w", lazy=True), help="Save the code here")
@click.pass_obj
async def get(obj, script):
    """
    Read a code entry.
    """
    if obj.meta:
        data, *meta = await obj.conn.d.get(obj.path)
        out = dict(data=data, meta=MsgMeta.restore(meta).repr())
    else:
        out = await obj.conn.d_get(obj.path)

    if script:
        if obj.meta:
            code = out["data"].pop("code", NotGiven)
        else:
            code = out.pop("code", NotGiven)
        if isinstance(code, str):
            print(code, file=script)

    yprint(out, stream=obj.stdout)


@cli.command("set")
@click.option("-a", "--async", "use_async", is_flag=True, help="Run this script async")
@click.option("-A", "--sync", "use_sync", is_flag=True, help="Run this script sync")
@click.option("-t", "--thread", is_flag=True, help="The code should run in a worker thread")
@click.option("-S", "--script", type=click.File(mode="r"), help="File with the code")
@click.option("-i", "--info", type=str, help="One-line description")
@click.option("-d", "--data", type=click.File(mode="r"), help="Load the metadata (YAML)")
@attr_args
@click.pass_obj
async def set_(obj, thread, script, data, use_async, use_sync, info, **kw):
    """
    Save Python code.
    """
    if use_async and use_sync:
        raise click.UsageError("--async and --sync are opposites.")
    if thread and (use_async or use_sync):
        raise click.UsageError("--thread conflicts with --async/--sync.")

    if data:
        msg = yload(data)
    else:
        msg = await obj.conn.d_get(obj.path)

    if thread:
        msg["is_async"] = False
    elif use_async:
        msg["is_async"] = True
    elif use_sync:
        msg["is_async"] = None
    if info is not None:
        msg["info"] = info

    if script:
        msg["code"] = script.read()
    elif "code" not in msg:
        raise click.UsageError("Missing script")

    vars_ = _sanitize_vars(msg.get("vars", {}))
    msg["vars"] = process_args(vars_, **kw)

    _check_exec_syntax(msg, obj.path)
    res = await obj.conn.d_set(obj.path, msg)
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command("list")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use for values.",
)
@click.option("-m", "--maxdepth", type=int, default=None, help="Limit recursion depth.")
@click.option("-M", "--mindepth", type=int, default=None, help="Starting depth.")
@click.option("-f", "--full", is_flag=True, help="Print complete entries.")
@click.option("-s", "--short", is_flag=True, help="Print shortened entries.")
@click.pass_obj
async def list_(obj, as_dict, maxdepth, mindepth, full, short):
    """
    List code entries.
    """
    await _list_entries(obj, as_dict, maxdepth, mindepth, full, short)


@cli.command()
@click.pass_obj
async def delete(obj):
    """
    Remove a code entry.
    """
    try:
        res = await obj.conn.d.delete(obj.path)
    except KeyError:
        if obj.debug:
            print("Does not exist.", file=obj.stdout)
        return

    if obj.meta:
        yprint(dict(data=res[0], meta=MsgMeta.restore(res[1:]).repr()), stream=obj.stdout)
    elif obj.debug:
        print("Deleted.", file=obj.stdout)


@cli.command(short_help="Edit an entry interactively")
@click.option("-e", "--editor", type=str, default=None, help="Editor (default: $EDITOR or vi)")
@click.pass_obj
async def edit(obj, editor):
    """
    Edit a code entry interactively.

    The first edit opens the complete YAML record. Follow-up actions let you
    edit either only ``code`` (``.py`` file), only non-code data
    (``.yaml`` file), or the full record.
    """
    if editor is None:
        editor = os.environ.get("VISUAL", os.environ.get("EDITOR", "vi"))

    try:
        original = await obj.conn.d_get(obj.path)
    except KeyError:
        try:
            original = await obj.conn.d_search(P("template") + obj.path)
        except KeyError:
            original = {"code": "return 42;\n"}
    current = dict(original)
    mode = EDIT_WHOLE

    while True:
        try:
            if mode == EDIT_WHOLE:
                current = await edit_yaml(editor, current)
            elif mode == EDIT_CODE:
                code = current.get("code", "")
                if code in (NotGiven, None):
                    code = ""
                if not isinstance(code, str):
                    raise click.UsageError("code must be a string")
                code = await edit_text(editor, code, suffix=".py")
                current["code"] = code
            elif mode == EDIT_NON_CODE:
                code = current.pop("code", NotGiven)
                try:
                    current = await edit_yaml(editor, current)
                finally:
                    if code is not NotGiven:
                        current["code"] = code
            else:
                raise RuntimeError(f"Invalid edit mode: {mode!r}")
            _check_exec_syntax(current, obj.path)
        except Exception as exc:
            click.echo(f"Edit failed: {exc}", err=True)
            mode = await click.prompt(
                "Continue with [w]hole, [c]ode, [n]on-code, or [a]bort?",
                type=click.Choice([EDIT_WHOLE, EDIT_CODE, EDIT_NON_CODE, EDIT_ABORT]),
                default=mode if mode in (EDIT_WHOLE, EDIT_CODE, EDIT_NON_CODE) else EDIT_WHOLE,
            )
        else:
            mode = await click.prompt(
                "Next: [w]hole, [c]ode, [n]on-code, [s]ave, [a]bort?",
                type=click.Choice([EDIT_WHOLE, EDIT_CODE, EDIT_NON_CODE, EDIT_SAVE, EDIT_ABORT]),
                default=EDIT_SAVE,
            )
        if mode == EDIT_ABORT:
            click.echo("Not saved.", err=True)
            return
        if mode == EDIT_SAVE:
            if current == original:
                click.echo("No changes.", err=True)
                return
            _check_exec_syntax(current, obj.path)
            res = await obj.conn.d_set(obj.path, current)
            if obj.meta:
                yprint(res, stream=obj.stdout)
            else:
                click.echo("Saved.", err=True)
            return



# command line interface
from __future__ import annotations

import anyio
import time
from contextlib import suppress

import asyncclick as click
from attrs import define, field

from moat.util import NotGiven, get_part, yprint
from moat.lib.path import P, Path
from moat.lib.priomap import TimerMap
from moat.lib.run import AliasedGroup
from moat.link.client import Link
from moat.link.node import Node

from collections.abc import Iterator, Mapping
from typing import Any, cast


@click.group(cls=AliasedGroup, short_help="Manage data flows.")
@click.pass_context
async def cli(ctx):
    """
    This subcommand reads data flow controls stored in the MoaT-Link service.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    obj.conn = await ctx.with_async_resource(Link(cfg, common=True))


def flow_path(path: Path) -> Path:
    """
    Return the flow-check lookup path for a data path.
    """
    return P("flow") + path


def _error_path(path: Path, rel: Path) -> Path:
    """
    Return the error path for a checked data sub-item.
    """
    if len(rel) == 0:
        return path
    return path / None + rel


def _iter_checks(data: Any, base: Path = Path()) -> Iterator[tuple[Path, Mapping[str, Any]]]:
    """
    Yield check definitions in a flow config tree.
    """
    if not isinstance(data, Mapping):
        return

    check = data.get("_")
    if isinstance(check, Mapping):
        yield base, check

    for key, value in data.items():
        if key == "_":
            continue
        if not isinstance(key, (str, int)):
            continue
        yield from _iter_checks(value, base / key)


def _as_float(value: Any) -> float | None:
    if isinstance(value, bool):
        return None
    if isinstance(value, (int, float)):
        return float(value)
    return None


def _check_limits(value: Any, check: Mapping[str, Any], previous: Any = NotGiven) -> str | None:
    """
    Validate min/max/maxstep limits.
    """
    num = _as_float(value)
    if (v_min := check.get("min", None)) is not None:
        if num is None:
            return "Value is not numeric"
        if num < float(v_min):
            return f"{num:g} < {float(v_min):g}"

    if (v_max := check.get("max", None)) is not None:
        if num is None:
            return "Value is not numeric"
        if num > float(v_max):
            return f"{num:g} > {float(v_max):g}"

    if (v_step := check.get("maxstep", None)) is not None and previous is not NotGiven:
        prev_num = _as_float(previous)
        if num is not None and prev_num is not None:
            delta = abs(num - prev_num)
            if delta > float(v_step):
                return f"Step {delta:g} > {float(v_step):g}"

    return None


async def _check_copied(conn: Link, value: Any, check: Mapping[str, Any]) -> str | None:
    """
    Validate a ``copied`` rule.
    """
    copied = check.get("copied", None)
    if not isinstance(copied, Mapping):
        return None

    if "at" not in copied:
        return "Copied rule has no destination path"

    try:
        c_val = await conn.d_get(copied["at"])
    except Exception as exc:
        return f"Copied lookup failed: {exc}"

    c_item = copied.get("item", None)
    if c_item is not None:
        try:
            c_val = get_part(c_val, c_item)
        except Exception as exc:
            return f"Copied item lookup failed: {exc}"

    if c_val != value:
        return f"Copied mismatch: want {value!r}, got {c_val!r}"
    return None


def _flow_for(flows: Any, path: Path) -> Mapping[str, Any] | None:
    try:
        flow = flows.search(path).data
    except (KeyError, ValueError):
        return None
    if not isinstance(flow, Mapping):
        return None
    return flow


async def _check_path(
    conn: Link,
    *,
    path: Path,
    data: Any,
    meta: Any,
    flow: Mapping[str, Any],
    previous: dict[Path, Any],
    now: float,
    do_copied: bool,
    do_stale_age: bool = True,
    queue_copied: bool = False,
) -> tuple[
    set[Path],
    dict[Path, tuple[str, Mapping[str, Any], Any]],
    list[_Timeout],
    list[_Copied],
]:
    """
    Evaluate one message against flow rules.
    """
    checked: set[Path] = set()
    errors: dict[Path, tuple[str, Mapping[str, Any], Any]] = {}
    timeouts: list[_Timeout] = []
    copied: list[_Copied] = []

    for rel, check in _iter_checks(flow):
        err_p = _error_path(path, rel)
        checked.add(err_p)

        if data is NotGiven:
            if check.get("required", False):
                errors[err_p] = ("Missing data", check, NotGiven)
            continue

        try:
            if len(rel) == 0:
                value = data
            else:
                value = get_part(data, rel)
        except Exception:
            value = NotGiven

        if value is NotGiven:
            if check.get("required", False):
                errors[err_p] = ("Missing data", check, NotGiven)
            continue

        prev_key = path + rel
        prev_val = previous.get(prev_key, NotGiven)
        previous[prev_key] = value

        msg = _check_limits(value, check, prev_val)
        if msg is None:
            if do_copied:
                msg = await _check_copied(conn, value, check)
            elif queue_copied:
                rule = check.get("copied", None)
                if isinstance(rule, Mapping):
                    delay = rule.get("delay", 3.0)
                    try:
                        delay = float(delay)
                    except (TypeError, ValueError):
                        delay = 3.0
                    if delay < 0:
                        delay = 0.0
                    copied.append(_Copied(err_p, delay, check, value))

        if msg is not None:
            errors[err_p] = (msg, check, value)
            continue

        timeout = check.get("timeout", None)
        if timeout is not None:
            try:
                timeout = float(timeout)
            except (TypeError, ValueError):
                timeout = None

        if timeout is not None and timeout > 0:
            timeouts.append(_Timeout(err_p, timeout, check, value, rel))
            ts = getattr(meta, "timestamp", None)
            if do_stale_age and ts is not None and now - ts > timeout and err_p not in errors:
                age = now - ts
                errors[err_p] = (f"Stale data: {age:g}s > {timeout:g}s", check, value)

    return checked, errors, timeouts, copied


@define
class _Timeout:
    path: Path = field()
    timeout: float = field()
    check: Mapping[str, Any] = field()
    value: Any = field()
    rel: Path = field()

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other: object):
        if isinstance(other, _Timeout):
            other = other.path
        if isinstance(other, Path):
            return self.path == other
        return False


@define
class _Copied:
    path: Path = field()
    delay: float = field()
    check: Mapping[str, Any] = field()
    value: Any = field()

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other: object):
        if isinstance(other, _Copied):
            other = other.path
        if isinstance(other, Path):
            return self.path == other
        return False


class FlowMon:
    """
    Monitor incoming messages and create flow errors.
    """

    def __init__(self, conn: Link, path: Path):
        self.conn = conn
        self.path = path
        self.previous: dict[Path, Any] = {}
        self.timeout: TimerMap[_Timeout] = TimerMap()
        self.copied: TimerMap[_Copied] = TimerMap()
        self.errs: Node = Node()
        self.flows: Node = Node()

    async def _set_error(self, path: Path, msg: str, check: Mapping[str, Any], data: Any) -> None:
        try:
            err = self.errs[path]
            if err.data.msg == msg:
                return
        except KeyError:
            pass
        await self.conn.e_info(
            P("flow") + self.path + path,
            msg,
            check=check,
            data=data,
        )

    async def _clear_error(self, path: Path) -> None:
        try:
            _err = self.errs[path]
        except KeyError:
            pass
        else:
            await self.conn.e_ok(P("flow") + self.path + path)

    async def _run_timeouts(self) -> None:
        async for tm in self.timeout:
            msg = f"No update for {tm.timeout:g}s"
            await self._set_error(tm.path, msg, tm.check, tm.value)

    async def _run_copied(self) -> None:
        async for cp in self.copied:
            msg = await _check_copied(self.conn, cp.value, cp.check)
            if msg is None:
                try:
                    if self.errs[cp.path].data.msg.startswith("Copied "):
                        await self._clear_error(cp.path)
                except (KeyError, ValueError):
                    pass
                continue
            await self._set_error(cp.path, msg, cp.check, cp.value)

    async def run(self) -> None:
        """
        Run the monitor.
        """
        async with (
            self.conn.d_watch(P("flow"), state=None, subtree=True).node as flows,
            self.conn.d_watch(P("error.flow") + self.path, state=None, subtree=True).node as errs,
            self.conn.d_watch(self.path, state=None, subtree=True, meta=True) as mon,
            anyio.create_task_group() as tg,
        ):
            self.flows = flows
            self.errs = errs
            tg.start_soon(self._run_timeouts)
            tg.start_soon(self._run_copied)

            async for rel, data, meta in mon:
                flow = _flow_for(flows, self.path + rel)
                if flow is None:
                    continue

                checked, errors, timeouts, copied = await _check_path(
                    self.conn,
                    path=rel,
                    data=data,
                    meta=meta,
                    flow=flow,
                    previous=self.previous,
                    now=time.time(),
                    do_copied=False,
                    do_stale_age=False,
                    queue_copied=True,
                )

                for path in checked:
                    if path not in timeouts:
                        with suppress(KeyError):
                            del self.timeout[cast(_Timeout, path)]

                    if path not in copied:
                        with suppress(KeyError):
                            del self.copied[cast(_Copied, path)]

                for tm in timeouts:
                    self.timeout[tm] = tm.timeout

                for cp in copied:
                    self.copied[cp] = cp.delay

                for path in checked:
                    if path in errors:
                        msg, check, value = errors[path]
                        await self._set_error(path, msg, check, value)
                    else:
                        await self._clear_error(path)


@cli.command()
@click.argument("path", type=P, nargs=1)
@click.pass_obj
async def get(obj, path):
    """
    Retrieve the flow-check definition for a path.
    """
    res = await obj.conn.d_search(flow_path(path))
    yprint(res, stream=obj.stdout)


@cli.command()
@click.argument("path", type=P, nargs=1, default=P("state"))
@click.pass_obj
async def monitor(obj, path):
    """
    Monitor a subtree and report flow errors.
    """
    flow = FlowMon(obj.conn, path)
    await flow.run()


@cli.command()
@click.option("-v", "--verbose", is_flag=True, help="Report incoming states.")
@click.option("-m", "--monitor", is_flag=True, help="Continue monitoring.")
@click.argument("path", type=P, nargs=1)
@click.pass_obj
async def check(obj, path, verbose, monitor):
    """
    Check stored messages in a subtree against flow checks.

    This command does not create errors in ``error.flow``.
    """
    n_bad = 0
    n_skip = 0
    n_good = 0
    previous: dict[Path, Any] = {}
    watch_state = None if monitor else True

    async with (
        obj.conn.d_watch(P("flow"), state=None, subtree=True).node as flows,
        obj.conn.d_watch(path, state=watch_state, subtree=True, meta=True) as mon,
    ):
        async for rel, data, meta in mon:
            full = path + rel
            flow = _flow_for(flows, full)
            if flow is None:
                n_skip += 1
                continue

            checked, errors, _timeouts, _copied = await _check_path(
                obj.conn,
                path=rel,
                data=data,
                meta=meta,
                flow=flow,
                previous=previous,
                now=time.time(),
                do_copied=True,
                queue_copied=False,
            )

            n_bad += len(errors)
            n_good += len(checked) - len(errors)
            for err_path, (detail, check, value) in errors.items():
                yprint(
                    dict(
                        path=path + err_path,
                        data_path=full,
                        detail=detail,
                        check=check,
                        data=value,
                    ),
                    stream=obj.stdout,
                )
                print("---", file=obj.stdout)
                obj.stdout.flush()

            if verbose:
                click.echo(f"ok: {n_good}  skipped: {n_skip}  bad: {n_bad}", err=True)



"""
MoaT-Link services announce themselves with defined names.
"""

from __future__ import annotations

import anyio
import logging
import os
import platform
import re
from anyio.abc import TaskStatus
from contextlib import asynccontextmanager

from attrs import define, field

from moat.util import NotGiven, attrdict, gen_ident
from moat.util import as_service as _as_service
from moat.lib.path import P, Path

from .client import Link
from .exceptions import ServiceCleared, ServiceNotFound, ServiceSupplanted

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import MsgSender

    from .client import LinkSender

    from typing import Any

__all__ = ["announcing"]

logger = logging.getLogger(__name__)


_unesc_re = re.compile(r"\\x([0-9a-fA-F]{2})")


def _unesc(s: PathElem) -> PathElem:
    if not isinstance(s, str):
        return s

    if "\\x" not in s:
        return s

    def _replace(match: re.Match[str]) -> str:
        return chr(int(match.group(1), 16))

    return _unesc_re.sub(_replace, s)


async def get_service_path(host: Path | str | bool, name: Path | None = None):
    path: list[PathElem] = ["run", "host"]

    match host:
        case True:
            path.append(platform.node())
        case False:
            pass
        case Path():
            path.extend(list(host.raw))
        case str():
            path.append(host)
        case _:
            raise TypeError("Host must be bool/path/str, not {host !r}.")

    if name is not None:
        path.extend(name.raw)
        return Path.build(path)

    if (fake := os.environ.get("MT_SERVICE", None)) is not None:
        path.extend(P(fake))
        return Path.build(path)

    if int(os.environ.get("SYSTEMD_EXEC_PID", "0")) == os.getpid():
        async with await anyio.Path("/proc/self/cgroup").open("r") as cgf:
            async for cg in cgf:
                logger.info("Control Group: %s", cg)
                for cge in cg.strip().split("/"):
                    if cge.endswith(".service"):
                        cge = cge[:-8]  # noqa:PLW2901
                        hi = cge.split("@", 1)
                        if hi[0] == "user":
                            continue  # ignore
                        if len(hi) != 1 or hi[0] != "moat-link-host":
                            for h in hi:
                                path.extend(h.split("-"))
                        logger.info("Control Path: %s", path)
                        return Path.build([_unesc(x) for x in path])
    raise ServiceNotFound


@define
class SetReady:
    """A handler to signal readiness.

    This duck-types `anyio.abc.TaskStatus` and `anyio.Event`.
    """

    link: LinkSender = field()
    path: Path = field()
    force: bool = field(default=False)
    service_path: Path | None = field(default=None)

    evt: anyio.Event = field(init=False, factory=anyio.Event)
    ready: anyio.Event | None = field(kw_only=True, default=None, repr=False)
    up: bool = field(init=False, default=False, repr=False)
    _value: Any = field(kw_only=True, default=None)

    def set(self):
        """
        We started up.
        """
        if self.up:
            raise RuntimeError("can only start once")
        self.up = True
        self.evt.set()

    def started(self, value: Any = None):
        """
        This mimics `TaskStatus.started`. The value, if any, is added to
        the announcement message as its ``value`` item.
        """
        if value is not None:
            self._value = value
        self.set()

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any):
        "Set the value to announce"
        self._value = value
        self.evt.set()

    async def monitor(self, *, task_status):
        """
        Watch for other announcements with a different ID.
        """
        async with self.link.d_watch(
            self.path, state=False if self.force else None, mark=True
        ) as mon:
            async for msg in mon:
                if msg is None:
                    if task_status is not None:
                        task_status.started()
                        task_status = None
                    continue

                if msg is NotGiven:
                    raise ServiceCleared(self.path)

                if msg["id"] != self.link.id and await self.link.i_checkid(msg["id"]):
                    raise ServiceSupplanted(self.path)

    async def wait_sr(self):
        await anyio.sleep(5)
        if not self.up:
            logger.warning(f"{self.path} did not (yet?) start")

    async def wait_evt(self, via: anyio.Event):
        await via.wait()
        self.set()

    async def run(self) -> None:
        """
        Send our announcement
        """
        while True:
            data: dict[str, Any] = {"id": self.link.id}
            if self.service_path:
                data["path"] = self.service_path

            # if it's already gone, don't worry.
            if self.value is not None:
                data["value"] = self.value
            if self.up is not None:
                data["up"] = self.up

            await self.link.d_set(self.path, data, retain=True)
            if self.ready is not None:
                await self.link.i_sync()
                self.ready.set()
                self.ready = None

            await self.evt.wait()
            self.evt = anyio.Event()


class FakeReady(TaskStatus):
    """A fake event/TaskStatus that doesn't do anything at all"""

    def set(self):
        pass

    def started(self, value: Any = None):
        pass


@asynccontextmanager
async def announcing(
    link: LinkSender,
    name: Path | None = None,
    *,
    host: Path | str | bool = True,
    force: bool = False,
    service: MsgSender | None = None,
    via: anyio.Event | None = None,
    value: Any = None,
):
    """
    This async context manager broadcasts the availability of a named service.

    Arguments:
        link: the link to use.
        host: a way to override the discovered service name.
        name: additional elements on the announcement
        service: Service to delegate to
        force: Flag to override an existing server
        via: Use this event as a readiness proxy.

    The context manager yields a SetReady object.

    Services are declared by a host prefix and a named path.

    ``host`` can be
    * ``True``: Use the hostname of the system the service runs on.
      This is the default.
    * a string: replace the hostname of the system the service runs on.
      Used for services that should run once in an installation.
    * ``False``: Don't add a hostname. This should not normally be done.

    If ``name`` is `None`, it will be auto-discovered from
    the current systemd unit, by selecting service entries in
    ``/proc/self/cgroups``. It is an error for the result to be empty.
    This requires the ``SYSTEMD_EXEC_PID`` environment variable to
    contain the PID of the current process.

    A :class:`~moat.link.exceptions.ServiceSupplanted` exception will be
    raised if there is a duplicate of the named service on the link
    (or one appears).

    The ``force`` flag is intended for replacing an existing service. It
    does not ignore a subsequent announcement by some other MoaT-Link
    client.
    """

    try:
        path = await get_service_path(host, name)
    except ServiceNotFound:
        logger.warning("Not running in a service CGroup. Set MT_SERVICE envvar?")
        yield FakeReady()
        return

    if path in link.announced:
        raise RuntimeError("A service announcement is already running")
    link.announced.add(path)

    service_path = Path.build((gen_ident(12),)) if service is not None else None

    async def _delegate(
        path: Path,
        service: MsgSender,
        *,
        task_status: TaskStatus[None] = anyio.TASK_STATUS_IGNORED,
    ) -> None:
        with link.link.delegate(path, service):
            task_status.started()
            await anyio.sleep_forever()

    async with anyio.create_task_group() as tg:
        try:
            if service:
                await tg.start(_delegate, service_path, service)
            srv = SetReady(link, path, force, service_path)
            if value is not None:
                srv.value = value
            await tg.start(srv.monitor)
            tg.start_soon(srv.run)
            tg.start_soon(srv.wait_sr)
            if via:
                tg.start_soon(srv.wait_evt, via)

            await link.i_sync()

            yield srv

            # only remove the host entry when we terminated normally.
            await link.d_set(path, retain=True)

        except ServiceSupplanted:
            logger.warning("Service %s on %s already exists", name, host)
            raise
        finally:
            link.announced.discard(path)
            tg.cancel_scope.cancel()


@asynccontextmanager
async def as_service(obj: attrdict | None = None):
    """
    This is a replacement for the legacy :func:`moat.util.as_service`
    helper that also registers the named service with MoaT-Link.
    """
    if obj is None:
        obj = attrdict()
    obj.setdefault("debug", False)

    async with (
        _as_service(obj) as mon,
        Link(obj.cfg.link, common=True) as mon.link,
        announcing(mon.link, via=mon.evt),
    ):
        yield mon



# command line interface  # noqa: D100
from __future__ import annotations

import asyncclick as click

from moat.util import yprint
from moat.lib.path import P
from moat.lib.run import attr_args, process_args
from moat.link.client import Link

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import MsgSender


@click.command(short_help="Send a command")
@click.option("-S", "--stream", is_flag=True, help="Read a stream")
@click.option("-R", "--raw", is_flag=True, help="Show raw message data")
@click.option("-C", "--client", type=P, help="Connect to this client")
@click.argument("path", type=P, nargs=1)
@click.pass_context
@attr_args(with_path=True, with_arglist=True)
async def cli(ctx, path, stream, raw, client, **kw):
    """
    Send a command to the server.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    if obj.get("port", None) is not None:
        cfg.client.port = obj.port

    async with Link(cfg) as conn:
        svc: MsgSender = await conn.get_service(client) if client else conn.sender

        kw = process_args({None: []}, no_path=True, **kw)
        args = kw.pop(None)
        if stream:
            async with svc.cmd(path, *args, **kw).stream_in() as res:
                yprint(rep_(res, raw), stream=obj.stdout)
                print("---", file=obj.stdout)
                async for msg in res:
                    yprint(rep_(msg, raw), stream=obj.stdout)
                    print("---", file=obj.stdout)
        else:
            res = await svc.cmd(path, *args, **kw)
        yprint(rep_(res, raw), stream=obj.stdout)


def rep_(res, raw=False):  # noqa: D103
    if raw:
        return res.args + [res.kw]
    if not res.args:
        return res.kw
    if res.kw:
        return res.args + [res.kw]
    if len(res.args) == 1:
        return res.args[0]
    return res.args



"""
The MoaT-Link client
"""

from __future__ import annotations

import anyio
import logging
import os
import platform
import sys
import time
from contextlib import asynccontextmanager, nullcontext, suppress
from contextvars import ContextVar
from functools import partial
from traceback import format_exception

import outcome
from attrs import define, field

from moat.util import (
    CtxObj,
    NotGiven,
    ValueEvent,
    attrdict,
    ctx_as,
    gen_ident,
    merge,
    srepr,
    timed_ctx,
    to_attrdict,
    ungroup,
)
from moat.lib.config import CFG
from moat.lib.mqtt import QoS, RetainHandling
from moat.lib.path import (
    P,
    Path,
    PathLongener,
    Root,
)
from moat.lib.rpc import Caller, MsgSender
from moat.util.random import al_unique

from .auth import AnonAuth, TokenAuth
from .common import CmdCommon
from .conn import TCPConn, UnixConn
from .exceptions import AuthError, ClientCancelledError
from .hello import Hello
from .meta import MsgMeta
from .node import Node

from typing import TYPE_CHECKING, overload

try:
    from .schema import schema_path, validate_instance
except ImportError:
    _HAVE_SCHEMA = False

    def schema_path(_path: Path) -> Path:
        raise RuntimeError("Schema support is not installed")

    def validate_instance(_schema: object, _data: object) -> None:
        raise RuntimeError("Schema support is not installed")

else:
    _HAVE_SCHEMA = True

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from moat.lib.path import PathElem
    from moat.lib.rpc import Msg
    from moat.link.code.run import Code
    from moat.link.node.codec import CodecNode

    from .auth import AuthMethod
    from .backend import Backend, Message

    from collections.abc import AsyncIterator, Awaitable, Callable
    from typing import Any, Literal, Protocol

    class DataSender(Protocol):
        "Protocol for `d.*` RPC sender helpers."

        get: Callable[[Path], Awaitable[Any]]
        search: Callable[[Path], Awaitable[Any]]
        set: Callable[[Path, Any, MsgMeta], Awaitable[Any]]
        walk: Callable[..., Caller]

    class ErrSender(Protocol):
        "Protocol for `e.*` RPC sender helpers."

        exc: Callable[..., Awaitable[Any]]
        info: Callable[..., Awaitable[Any]]
        ack: Callable[..., Awaitable[Any]]
        ok: Callable[..., Awaitable[Any]]
        mon: Callable[..., AbstractAsyncContextManager[Any]]


NotGivenType = type(NotGiven)


class _Requeue(Exception):
    pass


__all__ = [
    "BasicLink",
    "ClientCaller",
    "CodeCaller",
    "Link",
    "LinkCommon",
    "LinkSender",
    "get_link",
]

_the_link = ContextVar("_the_link")

local_addrs = {"localhost", "127.0.0.1", "::1"}


def get_link() -> Link | None:
    """
    Return the global Link instance, if one exists.
    """
    return _the_link.get(None)


class TS:
    "A wrapper to TaskStatus that swallows successive calls"

    def __init__(self, ts):
        self.ts = ts

    def started(self, data=None):
        "call .started once"
        self.ts.started(data)
        self.ts = anyio.TASK_STATUS_IGNORED


class BasicCmd:
    """
    A simple command that doesn't require streaming.

    The command can thus be repeated if a connection dies while it's
    running.
    """

    def __init__(self, a, kw):
        self.a = a
        self.kw = kw
        self._result = None
        self._evt = anyio.Event()

    async def run(self, link):
        "run the command"
        try:
            res = await link.cmd(*self.a, **self.kw)
        except anyio.get_cancelled_exc_class():
            raise
        except Exception as exc:
            self._result = outcome.Error(exc)
        else:
            self._result = outcome.Value(res)
        finally:
            self._evt.set()

    @property
    async def result(self):
        "retreve the result"
        await self._evt.wait()
        if self._result is None:
            raise _Requeue
        return self._result.unwrap()


class LinkCommon(CmdCommon):
    """
    Common code for links to MQTT only (BasicLink) and to a MoaT-Link server (Link).
    """

    protocol_version: int = -1
    name: str
    _id: str
    server_name: str | None = None
    is_server: bool = False
    _sender: MsgSender
    _hello: Hello | None = None
    _port: dict[str, Any] | str | None = None
    sdr: LinkSender
    announced: set[Path]
    backend: Backend
    tg: Any
    current_server: MsgSender | None

    def __init__(self, cfg, name: str | None = None):
        super().__init__(cfg)
        CFG.maybe_redo()

        if name is not None:
            self.is_server = True
        else:
            name = cfg.get("client_id")
            if name is None:
                name = "_" + gen_ident(12, alphabet=al_unique)
        self._id = name
        self.name = name

        self._cmdq_w, self._cmdq_r = anyio.create_memory_object_stream(5)
        self.logger = logging.getLogger(f"moat.link.client.{name}")
        self._sender = MsgSender(self)

    @property
    def sender(self) -> MsgSender:
        "The MsgSender that forwards to our server"
        return self._sender

    async def handle(self, msg, rcmd):
        """
        Message handler that intercepts incoming commands
        while authorization has not completed
        """
        if self._hello is not None and self._hello.auth_data is None:
            return await self._hello.handle(msg, rcmd)

        if rcmd and rcmd[-1] == "d_":
            msg._kw = dict(msg.kw)  # noqa: SLF001
            # reversed path, without the local "d_" suffix
            msg._kw["p"] = Path.build(rcmd[-2::-1])  # noqa: SLF001
            return await msg.call_stream(self.sdr.stream_d_)

        return await super().handle(msg, rcmd)

    @property
    def id(self):  # noqa:D102
        return self._id

    async def _connected_port(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        if self._port is None:
            raise RuntimeError("No configured port")
        async with self._connect_one(self._port) as hdl:
            task_status.started(hdl)
            await anyio.sleep_forever()

    def _state_changed(self):
        pass

    async def get_link(self) -> MsgSender:
        "Return the active remote sender."
        raise RuntimeError("No server link")

    @asynccontextmanager
    async def _connect_one(
        self, remote: dict[str, Any] | str, data: dict[str, Any] | None = None
    ) -> AsyncIterator[MsgSender]:
        auth_out: list[AuthMethod] = []
        rpc_auth_modes = ["anon"]
        rpc_auth_data: dict[str, Any] = {}
        if isinstance(remote, dict):
            if data is not None:
                with suppress(KeyError):
                    token = data["auth"]["token"]
                    auth_out.append(TokenAuth(token))
                    rpc_auth_modes.insert(0, "token")
                    rpc_auth_data["token"] = token
            conn_ = TCPConn(
                self,
                remote_host=remote["host"],
                remote_port=remote["port"],
                logger=self.logger,
            )
        else:
            conn_ = UnixConn(self, path=remote, logger=self.logger.debug)

        auth_out.append(AnonAuth())
        self._hello = Hello(
            me=self.name,
            me_server=self.is_server,
            auth_out=auth_out,
            rpc_auth_modes=tuple(rpc_auth_modes),
            rpc_auth_data=rpc_auth_data,
            rpc_auth_server=False,
        )
        yielded = False

        async with conn_ as conn:
            if self.logger.isEnabledFor(logging.INFO):
                self.logger.info("Connection %s to %s", self.name, srepr(remote))
            handler = MsgSender(conn)
            if (res := await self._hello.run(handler)) is False:
                raise AuthError("Initial handshake failed")

            if isinstance(res, dict):
                name = res.pop("name", None)
                if name is not None:
                    self.name = name
                if res:
                    self.logger.warning("Unknown auth reply: %r", res)

            if self._hello.me is not None:
                self.name = self._hello.me
            self.server_name = self._hello.them
            self.protocol_version = self._hello.protocol_version
            self._hello = None  # done with that
            self._state_changed()

            yielded = True
            yield handler

        if not yielded:
            raise EOFError


class ClientCaller(Caller):
    """
    This is a helper class that ensures the connection to the MoaT-Link server
    is up before attempting to talk to it.

    Don't instantiate this class yourself.
    """

    def __init__(self, sender, *a, **kw):
        self._link = sender._link  # noqa:SLF001
        super().__init__(sender, *a, **kw)

    async def _call(self):
        "helper for __await__ that calls the remote handler"
        link = await self._link.get_link()
        cmd, a, kw = self.data
        return await link.root._sender.cmd(cmd, *a, **kw)  # noqa:SLF001


@define
class _CodeWatch:
    """
    Shared watcher state for one code path.
    """

    code: Code
    users: int
    cancel_scope: anyio.CancelScope
    done: anyio.Event


def _code_path(path: Path) -> Path:
    "Build a code storage path below ``code.exec``."
    from moat.link.code import CODE_EXEC_ROOT  # noqa: PLC0415

    return CODE_EXEC_ROOT + Path.build(path)


class CodeCaller:
    """
    Helper returned by :meth:`LinkSender.code_at`.

    This object supports direct calls, ``await`` to get the current
    code wrapper, and async context usage to
    keep one shared wrapper updated in the background.
    """

    sender: LinkSender
    path: Path

    def __init__(self, sender: LinkSender, path: Path):
        self.sender = sender
        self.path = _code_path(path)

    def __await__(self):
        return self.sender._code_get(self.path).__await__()  # noqa: SLF001

    def __call__(self, **kw: Any):
        "Call the referenced code snippet."

        async def _run():
            code = await self
            return await code(**kw)

        return _run()

    async def _enter(self) -> Code:
        "Acquire (and maybe start) one shared code watcher."
        p = self.path
        while True:
            entry = self.sender._code_watch.get(p, NotGiven)  # noqa: SLF001
            if isinstance(entry, _CodeWatch):
                entry.users += 1
                return entry.code
            if entry is not NotGiven:
                await entry.wait()
                continue

            evt = anyio.Event()
            self.sender._code_watch[p] = evt  # noqa: SLF001
            try:
                from moat.link.code.run import Code  # noqa: PLC0415

                code = Code(self.sender, p)
                code.update(await self.sender.d_get(p))
                done = anyio.Event()
                scope = await self.sender._link.tg.start(self.sender._watch_code, p, code, done)  # noqa: SLF001
                self.sender._code_watch[p] = _CodeWatch(  # noqa: SLF001
                    code=code,
                    users=1,
                    cancel_scope=scope,
                    done=done,
                )
                evt.set()
                return code
            except BaseException:
                self.sender._code_watch.pop(p, None)  # noqa: SLF001
                evt.set()
                raise

    async def _exit(self) -> None:
        "Release one shared code watcher."
        p = self.path
        while True:
            entry = self.sender._code_watch.get(p, NotGiven)  # noqa: SLF001
            if isinstance(entry, _CodeWatch):
                if entry.users > 1:
                    entry.users -= 1
                    return
                evt = anyio.Event()
                self.sender._code_watch[p] = evt  # noqa: SLF001
                try:
                    entry.cancel_scope.cancel()
                    await entry.done.wait()
                finally:
                    self.sender._code_watch.pop(p, None)  # noqa: SLF001
                    evt.set()
                return
            if entry is not NotGiven:
                await entry.wait()
                continue
            if entry is NotGiven:
                return

    async def __aenter__(self) -> Code:
        return await self._enter()

    async def __aexit__(self, *tb):
        await self._exit()


class LinkSender(MsgSender):
    """
    This is the client-side front-end to a "standard" MoaT-Link connection.

    It contains various somewhat-high-level helpers. Most are named like
    the corresponding backend commands, with an underscore replacing the
    dot.

    Do not instantiate this class yourself; it's returned when you enter
    the context of a `Link`.
    """

    Caller_ = ClientCaller

    _codec_tree: CodecNode | None = None
    _codec_tree_wait: anyio.Event | None = None
    _code_watch: dict[Path, _CodeWatch | anyio.Event]
    d: DataSender
    d_: MsgSender
    e: ErrSender
    i: MsgSender
    cl: MsgSender

    def __init__(self, link: LinkCommon):
        self._link = link
        self.announced = self._link.announced
        self._code_watch = {}
        self._codec_tree_wait = None

    @property
    def root(self):
        "Self"
        return self

    @property
    def link(self):
        "Link"
        return self._link

    @property
    def name(self) -> str:
        "Link name"
        return self._link.name

    @property
    def cfg(self) -> attrdict:
        "Link config"
        return self._link.cfg

    @property
    def id(self) -> str:
        "Link ID"
        return self._link.id

    async def handle(self, msg: Msg, rcmd: list[PathElem]) -> None:
        """
        Standard handler, forwards to the remote side.
        """
        if rcmd and rcmd[-1] == "d_":
            return await self._link.handle(msg, rcmd)
        srv = await self._link.get_link()
        await srv.handle(msg, rcmd)

    def find_handler(self, path: Path, cmd: bool = False) -> tuple[MsgSender, Path]:
        """
        Standard sub-dispatcher redirector, no-op.
        """
        cmd  # noqa:B018
        return self, path

    @asynccontextmanager
    async def announcing(self, *a, **kw):
        """
        Frontend for `moat.link.announce.announcing`.
        """
        from moat.link.announce import announcing as ann  # noqa:PLC0415

        async with ann(self, *a, **kw) as res:
            yield res

    async def stream_d_(self, msg: Msg):
        """
        Simple Data Protocol on the client, possibly with subpath.
        """
        try:
            p = msg["p"]
        except KeyError:
            p = msg[0]
            off = 1
        else:
            off = 0

        p = Path.build(p)

        try:
            try:
                d = msg["d"]
            except KeyError:
                d = msg[off]
            # write

            await self.d_set(p, d)

        except (KeyError, IndexError):
            # read
            res = await self.d_get(p)
            await msg.result(res)

    @overload
    async def d_get(self, path: Path, meta: Literal[True]) -> tuple[Any, MsgMeta]: ...

    @overload
    async def d_get(self, path: Path) -> Any: ...

    async def d_get(self, path: Path, meta: bool = False) -> tuple[Any, MsgMeta]:
        """
        Data retrieval. Calls the server's ``d.get`` method.

        Returns a data+metadata tuple if @meta is True, otherwise just the
        data.
        """
        if len(path) and isinstance(path[0], Path):
            raise ValueError("Don't use a root-prefixed path here.")

        res = await self.d.get(path)
        if not meta:
            return res[0]
        return res[0], MsgMeta.restore(res[1:])

    @overload
    async def d_search(self, path: Path, meta: Literal[True]) -> tuple[Any, MsgMeta]: ...

    @overload
    async def d_search(self, path: Path) -> Any: ...

    async def d_search(self, path: Path, meta: bool = False) -> tuple[Any, MsgMeta]:
        """
        Search-based data retrieval. Calls the server's ``d.search`` method.

        Returns a data+metadata tuple if @meta is True, otherwise just the
        data.
        """
        if len(path) and isinstance(path[0], Path):
            raise ValueError("Don't use a root-prefixed path here.")

        res = await self.d.search(path)
        if not meta:
            return res[0]
        return res[0], MsgMeta.restore(res[1:])

    @overload
    async def d_set(
        self,
        path: Path,
        data: Any = NotGiven,
        t: float | None = None,
        meta: Literal[True] = True,
        retain: bool | None = None,
        verify: bool | None = False,
    ) -> bool | None: ...

    @overload
    async def d_set(
        self,
        path: Path,
        data: Any = NotGiven,
        t: float | None = None,
        meta: Literal[False] = False,
        retain: bool | None = None,
        verify: bool | None = False,
    ) -> bool | None: ...

    async def d_set(
        self,
        path: Path,
        data: Any = NotGiven,
        meta: MsgMeta | None = None,
        t: float | None = None,
        with_prev: bool = False,
        retain: bool | None = None,
        verify: bool | None | NotGivenType = NotGiven,
    ) -> bool | None | tuple[Any, MsgMeta | None]:
        """
        Data update.

        If a timestamp is passed in @t or the old value+metadata is
        requested via @with_prev, goes through the server. Otherwise posts to
        MQTT directly.
        """
        if path and isinstance(path[0], Path):
            raise ValueError("Don't use a root-prefixed path here.")

        if verify is NotGiven:
            verify = (
                "pytest" in sys.modules and getattr(self._link, "current_server", None) is not None
            )

        if verify is not False:
            if not _HAVE_SCHEMA:
                if verify is None:
                    self._link.logger.warning(
                        "Schema support is not installed, cannot verify %s", path
                    )
                else:
                    raise RuntimeError("Schema support is not installed")
            else:
                try:
                    schema = await self.d_search(schema_path(path))
                except KeyError:
                    pass
                except Exception as exc:
                    if verify is None:
                        self._link.logger.warning("Schema lookup failed for %s: %r", path, exc)
                    else:
                        raise
                else:
                    try:
                        validate_instance(schema, data)
                    except Exception as exc:
                        if verify is None:
                            self._link.logger.warning(
                                "Schema validation failed for %s: %r", path, exc
                            )
                        else:
                            raise

        if meta is None:
            meta = MsgMeta(origin=self._link.name)
        if t is None and not with_prev:
            if retain is None:
                retain = len(path) == 0 or path[0] != "run"
            await self.send(Root.get() + path, data=data, meta=meta, retain=retain)
            return True
        tt = {} if t is None else {"t": t}
        old: tuple[Any, MsgMeta | None] | None = None
        if with_prev:
            try:
                old = await self.d_get(path, meta=True)
            except KeyError:
                old = (NotGiven, None)
        res = await self.d.set(path, data, meta, **tt)
        if not isinstance(res, bool) and res is not None:
            res = res[0]
        if with_prev:
            assert old is not None
            return old
        return res

    @asynccontextmanager
    async def d_walk(
        self,
        path: Path,
        meta: bool = False,
        min_ts: float = 0,
        min_depth: int = 0,
        max_depth: int = 255,
    ) -> AsyncIterator[tuple[str, Any, MsgMeta]]:
        """
        Fetch a (limited) subtree.
        """
        args = [path, min_ts, min_depth, max_depth]
        if args[-1] == 255:
            args.pop()
            if args[-1] == 0:
                args.pop()
                if args[-1] == 0:
                    args.pop()

        async with self.d.walk(*args).stream_in() as mon:
            yield Walker(mon, meta=meta)

    async def stream_watch(self, msg: Msg):
        """
        A hook for reading data. Used mainly by `moat.lib.rpc.app.link.Cmd`.
        """
        if msg.can_stream:
            async with msg.stream_out() as ms, self.d_watch(msg[0], **msg.kw) as mr:
                async for d in mr:
                    await ms.send(d)
        else:
            d = await self.d_get(msg[0], **msg.kw)
            await msg.result(d)

    # No use marking as Any|None if @mark is set
    @overload
    def d_watch(
        self,
        path: Path,
        mark: bool = False,
        meta: Literal[False] = False,
        subtree: Literal[False] = False,
        state: bool | NotGivenType | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[Any]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[False] = False,
        meta: Literal[True] = True,
        subtree: Literal[False] = False,
        state: bool | NotGivenType | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[tuple[Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[False],
        meta: Literal[True],
        subtree: Literal[True],
        state: bool | NotGivenType | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[tuple[Path, Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[False] = False,
        meta: Literal[False] = False,
        subtree: Literal[True] = True,
        state: bool | NotGivenType | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[tuple[Path, Any]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[True],
        meta: Literal[True],
        subtree: Literal[False] = False,
        state: bool | NotGivenType | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[None | tuple[Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[True],
        meta: Literal[True],
        subtree: Literal[True],
        state: bool | NotGivenType | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[None | tuple[Path, Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[True],
        meta: Literal[False] = False,
        subtree: Literal[True] = True,
        state: bool | NotGivenType | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[None | tuple[Path, Any]]]: ...

    def d_watch(
        self,
        path: Path,
        meta: bool = False,
        subtree: bool = False,
        state: bool | None | NotGivenType = None,
        age: float | None = None,
        mark: bool = False,
        min_length: int | None = None,
        max_length: int | None = None,
        cls: type[Node] = Node,
    ):
        """
        Monitor a node or subtree.

        Args:
            path: what to examine
            meta: flag whether to return metadata too
            subtree: flag whether to watch a subtree, not just this node
            mark: yield `None` when the initial state has been transmitted
            state: send the current state (True), updates (False), both with
                    current data from the server (None; this is the
                    default), or both via MQTT (NotGiven).
            age: skip nodes with data older than this many seconds.
            cls: type of root node (default `~moat.link.node.Node`)

        This method returns an async context manager which yields an async iterator.
        The iterator yields node data if neither ``subtree`` nor ``meta`` is set.
        Otherwise it yields tuples. The first item is the path if ``subtree``
        is set; the last item is the metadata if ``meta`` is set.

        ``state=NotGiven`` does not work with ``mark=True``, or when the same
        subscription is already active somewhere else.
        """
        return Watcher(self, path, meta, subtree, state, age, mark, cls, min_length, max_length)

    @asynccontextmanager
    async def as_dict(self, path: Path, default: attrdict | None = None):
        """
        This async context manager returns an attrdict that gets auto-updated
        when a configuration tree changes.

        The root also contains a ``wait_`` member which can be awaited,
        for re-sync after updates.

        Warning: The order of interpreting these values and their updates
        is not specified. Conflicts should be strictly avoided.
        """
        res = attrdict() if default is None else default
        evt = anyio.Event()

        class AWT:
            def __await__(self):
                return evt.wait().__await__()

        res.wait_ = AWT()

        async with (
            anyio.create_task_group() as tg,
            Watcher(self, path, False, True, None, None, True, Node, None, None) as mon,
        ):

            async def _run(task_status):
                nonlocal evt

                async for pd in mon:
                    if pd is None:
                        evt = anyio.Event()
                        task_status.started()
                        continue
                    p, d = pd
                    if type(d) is dict:
                        d = to_attrdict(d)
                    dl = list(p)
                    while dl:
                        n = dl.pop()
                        d = attrdict({n: d})
                    if not isinstance(d, dict):
                        self._link.logger.warning("Item at %r is %r, not a dict, ignoring", p, d)
                        continue
                    merge(res, d)
                    if evt is not None:
                        evt.set()
                        evt = anyio.Event()

            await tg.start(_run)
            yield res
            tg.cancel_scope.cancel()

    async def e_exc(self, path: Path, exc: Exception, **kw):
        """
        Report an exception, with backtrace.
        """
        exc = ungroup.one(exc)

        kw["bt"] = format_exception(exc)
        kw["exc"] = exc

        await self.e.exc(path, repr(exc), **kw)

    async def e_info(self, path: Path, txt: str, **kw):
        """
        Report a problem that's not an exception.
        """
        await self.e.info(path, txt, **kw)

    async def e_ack(self, path: Path, **kw):
        """
        Report that $USER has been notified of a problem.

        @ack can be True (don't notify any more), False (notify next
        time it happens), or a timestamp (notify if it happens again after
        $time).
        """
        await self.e.ack(path, **kw)

    async def e_ok(self, path: Path, **kw):
        """
        Report that something is working.
        """
        await self.e.ok(path, **kw)

    @asynccontextmanager
    async def e_wrap(self, path: Path, **kw):
        """
        An error-handling context manager. It opens a stream to the server
        and reports either success or failure over it.

        Streamed messages can be sent and will be stored as auxiliary
        information if the command should fail.

        This wrapper does *not* swallow the exception.
        """
        async with self.e.mon(path, **kw) as mon:
            yield mon
            await mon.send(True)

    def monitor(self, path: Path, *a, **kw) -> AbstractAsyncContextManager[AsyncIterator[Message]]:
        """
        Watch this path.

        This call forwards directly to the back-end.
        See `moat.link.backend.Backend.monitor` for call details.

        The caller is responsible for prefixing the Root path, if applicable.
        """
        return self._link.backend.monitor(path, *a, **kw)

    def send(self, path: Path, *a, **kw) -> Awaitable[None]:
        """
        Send to this path.

        This call forwards directly to the back-end.
        See `moat.link.backend.Backend.send` for call details

        The caller is responsible for prefixing the Root path, if applicable.
        """
        return self._link.backend.send(path, *a, **kw)

    async def i_sync(self):
        """
        This call tries to ensure that the server has processed all
        incoming MQTT messages.

        It does this by requesting a timestamp from the server, publishing
        the timestamp to MQTT, and then waiting for the server to
        acknowledge having received (at least) this number.

        This code assumes that MQTT messages are delivered in the order
        they are submitted. Depending on the server's multithreading setup
        this may or may not be true, but in practice it should be
        sufficiently true-ish to work out.

        It also assumes that if there's more than one server, they all have
        mostly-accurate time.
        """
        (st,) = await self.cmd(P("i.stamp"))
        await self.send(P(":R.run.service.main.stamp") / self._link.server_name, st, retain=False)
        await self.cmd(P("i.sync"), st)

    async def i_checkid(self, id: str) -> bool:
        """
        Check whether the given client ID is still valid.
        """
        (st,) = await self.cmd(P("i.checkid"), id)
        return st

    async def _get_tree(self, path, *, task_status, **kw):
        async with Watcher(
            self,
            path,
            bool(kw.get("meta", False)),
            bool(kw.get("subtree", False)),
            kw.get("state", None),
            kw.get("age", None),
            bool(kw.get("mark", False)),
            kw.get("cls", Node),
            kw.get("min_length", None),
            kw.get("max_length", None),
        ) as w:
            task_status.started(await w.get_node())
            await anyio.sleep_forever()

    async def get_codec_tree(self):
        """
        Returns a common subtree for codecs.
        """

        if self._codec_tree is None:
            if self._codec_tree_wait is not None:
                await self._codec_tree_wait.wait()
                assert self._codec_tree is not None
                return self._codec_tree
            evt = anyio.Event()
            self._codec_tree_wait = evt
            from moat.link.node.codec import CodecNode  # noqa: PLC0415

            tree = await self._link.tg.start(
                partial(
                    self._get_tree, P("codec"), subtree=True, state=None, meta=False, cls=CodecNode
                )
            )
            self._codec_tree = tree
            self._codec_tree_wait = None
            evt.set()
        assert self._codec_tree is not None
        return self._codec_tree

    async def get_service(self, srv: Path, wait: bool = False) -> MsgSender:
        """
        Retrieve a SubDispatcher that connects to the given announced
        service.
        """

        srv = P("run.host") + srv
        try:
            res = await self.d_get(srv)
        except KeyError:
            if not wait:
                raise
            async with self.d_watch(srv) as mon:
                async for res in mon:
                    if res not in (None, NotGiven):
                        break

        res2 = await self.d_get(P("run.id") / res["id"])

        return self.sub_at(Path.build(("srv", res2["srv"], "cl", res["id"])) + res["path"])

    async def _code_get(self, p: Path) -> Code:
        "Return cached code while monitored, otherwise a fresh wrapper."
        while True:
            entry = self._code_watch.get(p, NotGiven)
            if isinstance(entry, _CodeWatch):
                return entry.code
            if entry is not NotGiven:
                await entry.wait()
                continue
            if entry is NotGiven:
                from moat.link.code.run import Code  # noqa: PLC0415

                return Code(self, p)

    async def _watch_code(
        self, path: Path, code: Code, done: anyio.Event, *, task_status=anyio.TASK_STATUS_IGNORED
    ):
        "Background updater for one code path."
        with anyio.CancelScope() as scope:
            task_status.started(scope)
            try:
                async with self.d_watch(path, state=False) as mon:
                    async for data in mon:
                        code.update(data)
            finally:
                done.set()

    def code_at(self, path: Path) -> CodeCaller:
        """
        Return a code caller.

        The result can be called directly, awaited to retrieve a
        code wrapper, or used as an async
        context manager to keep one shared wrapper updated.
        """
        return CodeCaller(self, path)


class Link(LinkCommon, CtxObj):
    """
    This class combines the back-end link with a connection to a MoaT-Link server.

    If "common" is set, create global state.
    If "only" is set, connect only to the named server.
    """

    _server: ValueEvent = None
    _uptodate: bool = False
    _hello: Hello | None = None
    current_server: MsgSender | None = None
    _server_up: anyio.Event
    _last_link: Message | None = None
    _last_link_seen: anyio.Event
    _port: dict[str, Any] | str | None = None
    _state: str = "init"
    _common: bool = False
    _only: str | None = None
    announced: set[Path]
    sdr: LinkSender

    def __new__(cls, cfg, name: str | None = None, common: bool = False, only: str | None = None):  # noqa:D102
        cfg, name, only  # noqa:B018
        if common and (link := _the_link.get(NotGiven)) is not NotGiven:
            return link
        return super().__new__(cls)

    def __init__(
        self, cfg, name: str | None = None, common: bool = False, only: str | None = None
    ):
        if _the_link.get(NotGiven) is self:
            return
        super().__init__(cfg, name=name)
        self._retry_msgs: set[BasicCmd] = set()
        self._server_up = anyio.Event()
        self._state_change = anyio.Event()
        self._common = common
        self._only = only
        self.announced = set()
        with suppress(AttributeError):
            self._port = self.cfg.client.port

    async def set_state(self, state: str):
        """
        Set the state string for our ping message
        """
        if state != self._state:
            self._state = state
            self._state_changed()

    def _state_changed(self):
        self._state_change.set()

    async def get_link(self):
        """
        This method refreshes the link to the server if it happens to be down.
        """
        while self.current_server is None:
            await self._server_up.wait()
        return self.current_server

    @property
    def _ping_path(self):
        return P("run.ping.id") / self.id

    @property
    def _id_path(self):
        return P("run.id") / self.id

    async def _send_id(self):
        data = dict(
            host=platform.node(),
            pid=os.getpid(),
            argv=sys.argv,
        )
        if self.server_name is not None:
            data["srv"] = self.server_name
        await self.sdr.d_set(
            self._id_path,
            data,
            retain=True,
        )

    async def _send_ping(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        Periodially (or when our state changes) send a Ping.

        This also sends our ID message, after the first ping,
        but only if the task status isn't ignored.
        """
        path = Root.get() + self._ping_path
        sn = self.server_name
        while True:
            await self.backend.send(
                path,
                data=dict(
                    up=True,
                    state=self._state,
                ),
                retain=False,
                meta=False,
            )

            if task_status is not anyio.TASK_STATUS_IGNORED:
                # send initial run.id message, *after* the ping
                sn = self.server_name
                await self._send_id()
                task_status.started()
                task_status = anyio.TASK_STATUS_IGNORED

            elif sn != self.server_name:
                sn = self.server_name
                await self._send_id()

            if self._state == "init":
                self._state = "auto"

            with anyio.move_on_after(self.cfg.timeout.ping.every):
                await self._state_change.wait()
                self._state_change = anyio.Event()

    async def _monitor_ping(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        path = Root.get() + self._ping_path
        async with self.backend.monitor(path) as mon:
            task_status.started()
            async for msg in mon:
                if msg.data is NotGiven:
                    raise ClientCancelledError(msg.meta)
                # anything else shall be ignored

    @asynccontextmanager
    async def _ctx(self):
        from .backend import get_backend  # noqa: PLC0415

        # clears our announcement on disconnect
        will = attrdict(
            data=dict(up=False, state="will"),
            topic=P(":R") + self._ping_path,
            retain=False,
            qos=QoS.AT_LEAST_ONCE,
        )
        async with (
            ctx_as(Root, self.cfg["root"]),
            get_backend(self.cfg, name=self.name, will=will) as self.backend,
        ):
            try:
                async with anyio.create_task_group() as self.tg:
                    if self._port is not None:
                        sdr = await self.tg.start(self._connected_port)
                    else:
                        if self.cfg.client.init_timeout:
                            # connect to the main server
                            await self.tg.start(self._run_server_link)
                        sdr = LinkSender(self)
                    sdr.add_sub("cl")
                    sdr.add_sub("d")
                    sdr.add_sub("d_")
                    sdr.add_sub("e")
                    sdr.add_sub("i")
                    try:
                        self.sdr = sdr
                        await self.tg.start(self._monitor_ping)
                        await self.tg.start(self._send_ping)

                        with ctx_as(_the_link, self) if self._common else nullcontext():
                            yield sdr
                    finally:
                        del self.sdr
                    self.tg.cancel_scope.cancel()

            finally:
                try:
                    with anyio.move_on_after(2, shield=True):
                        await self.backend.send(
                            Root.get() + self._ping_path,
                            data=dict(up=False, state="closed"),
                            retain=False,
                            meta=False,
                        )
                except Exception as exc:
                    self.logger.warning("Could not send Close message", exc_info=exc)

    def cancel(self):
        "Stop me"
        self.tg.cancel_scope.cancel()

    async def _run_server_link(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        This is the manager task for the server link channel.
        It starts a server connection (and tries to keep it alive).
        """
        task_status = TS(task_status)

        with anyio.fail_after(self.cfg.client.init_timeout):
            srv = await self.tg.start(self._read_server_link)

        tm = self.cfg.timeout.connect
        timeout = tm.initial
        while True:
            if srv is None:
                await self._last_link_seen.wait()
                srv = self._last_link
                self._last_link_seen = anyio.Event()
                continue
            try:
                await self._connect_server(srv, task_status=task_status)
            except Exception as exc:
                if srv.meta is None:
                    err_path = P("run.service.main.server")
                else:
                    err_path = P("run.service.main.server") / srv.meta.origin
                await self.backend.send_error(
                    err_path,
                    data=srv,
                    exc=exc,
                )
            finally:
                self.current_server = None
                if self._server_up.is_set():
                    self._server_up = anyio.Event()

                # TODO test if tasks from the remote side already get cancelled
                # TODO and if not, do that

            # Back off to either a new server announcement or the exponential timeout
            try:
                with anyio.fail_after(timeout):
                    await self._last_link_seen.wait()
            except TimeoutError:
                # try the last-reported connection again
                timeout = min(timeout * tm.factor, tm.max)
            else:
                # immediately use the new data
                srv = self._last_link
                self._last_link_seen = anyio.Event()

    async def _read_server_link(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        # Background process that monitors the server link channel

        self._last_link_seen = anyio.Event()
        if self._only is not None:
            # Monitor the specific named server's announcement
            monitor_path = P(":R.run.service.main.server") / self._only
        else:
            # Monitor the main connection channel
            monitor_path = P(":R.run.service.main.conn")
        async with self.backend.monitor(
            monitor_path,
            retain_handling=RetainHandling.SEND_RETAINED,
        ) as mon:
            async for msg in mon:
                self._last_link = msg
                if task_status is None:
                    self._last_link_seen.set()
                else:
                    task_status.started(msg)
                    task_status = None

    async def _connect_run(self, rem, *, task_status=anyio.TASK_STATUS_IGNORED):
        # We're connected.
        self.current_server = rem
        self._server_up.set()

        # Check backend compatibility when connecting to a named server
        if self._only is not None:
            try:
                remote_backend = await rem.cmd(P("i.backend"))
                if remote_backend.kw != self.cfg.backend:
                    self.logger.info(
                        "Server %s uses different backend: %r (we use %r)",
                        self._only,
                        remote_backend.kw,
                        self.cfg.backend,
                    )
                    # Store the remote backend for potential reconnection
                    self._remote_backend = remote_backend.kw
            except Exception as exc:
                self.logger.warning("Could not query backend from %s: %r", self._only, exc)

        async def run_(cmd):
            await cmd.run(rem)

        async with anyio.create_task_group() as tg:
            for msg in self._retry_msgs:
                tg.start_soon(run_, msg)
            task_status.started()
            async for msg in self._cmdq_r:
                tg.start_soon(run_, msg)

    async def cmd(self, *a, _idem=False, **kw):
        """
        Queue and run a simple command.

        If @_idem is False, the command will error out if the connection
        dies while it's running. Otherwise (the default) it may be repeated.
        """
        cmd_ = BasicCmd(a, kw)
        try:
            self._retry_msgs.add(cmd_)
            while True:
                while self.current_server is None:
                    await self._server_up.wait()

                try:
                    with ungroup:
                        await self._cmdq_w.send(cmd_)
                        return await cmd_.result
                except _Requeue:
                    if _idem:
                        raise EOFError from None

        finally:
            self._retry_msgs.discard(cmd_)

    async def _connect_server(self, srv: Message, *, task_status=anyio.TASK_STATUS_IGNORED):
        task_status = TS(task_status)

        # Backend connection
        link = srv.data["link"]
        if isinstance(link, dict):
            link = (link,)

        local_ok = srv.data.get("node", "localhost")

        for remote in link:
            if local_ok:
                pass
            elif "host" not in remote:
                continue
            elif remote["host"] in local_addrs:
                continue
            try:
                async with timed_ctx(
                    self.cfg.client.init_timeout, self._connect_one(remote, srv.data)
                ) as rem:
                    await self._connect_run(rem, task_status=task_status)
            except OSError as exc:
                self.logger.warning("Link failed: %r (%r)", remote, exc)
            except Exception as exc:
                self.logger.warning("Link failed: %r", remote, exc_info=exc)


class BasicLink(LinkCommon, CtxObj):
    """
    Simple direct link to a MQTT server.
    """

    def __init__(self, cfg, name: str | None, data: dict, **kw):
        super().__init__(cfg, name=name, **kw)
        self.data = data

    @asynccontextmanager
    async def _ctx(self):
        link = self.data["link"]
        if isinstance(link, dict):
            link = (link,)

        err = None
        yielded = False
        for remote in link:
            try:
                async with self._connect_one(remote, self.data) as self._sender:
                    yielded = True
                    yield self._sender
                    return
            except Exception as exc:
                if yielded:
                    raise
                self.logger.warning(
                    "Link failed: %r",
                    remote,
                    exc_info=None if isinstance(exc, EnvironmentError) else exc,
                )
                if err is None:
                    err = exc
            else:
                if not yielded and err is None:
                    err = EOFError

        if not yielded:
            if err is None:
                raise ValueError(f"No links in {self.data!r}")
            raise err


@define(eq=False)
class Watcher(CtxObj):
    """
    Helper class for monitoring (and coalescint the data of) either-or-both
    * a MQTT subscription to a subtree of our MoaT-Link hierarchy
    * a MoaT-Link request to enumerate a subtree
    """

    link: LinkSender = field()
    path: Path = field()
    meta: bool = field()
    subtree: bool = field()
    state: bool | None | NotGivenType = field()
    age: float | None = field()
    mark: bool = field()
    node_cls: type[Node] = field()
    min_length: int | None = field()
    max_length: int | None = field()

    _qr = field(init=False, repr=False)
    _tg = field(init=False, repr=False)
    _node = field(init=False, default=None)

    _current_done: anyio.Event | None = field(init=False, default=None)

    def __attrs_post_init__(self):
        if self.mark and self.state is NotGiven:
            raise ValueError("MQTT doesn't send a mark. Sorry.")
        if self.age is not None:
            self.age = time.time() - self.age

    def _chk(self, p, m):
        if self.age and m.timestamp < self.age:
            return False
        if self.min_length is not None and len(p) < self.min_length:
            return False
        if self.max_length is not None and len(p) > self.max_length:
            return False
        return True

    async def _current(self, qw, *, task_status):
        "get the current state from the server"
        if self.subtree:
            pl = PathLongener(())
            args = [self.path, self.age, self.min_length, self.max_length]
            while args[-1] is None:
                args.pop()
            async with self.link.d.walk(*args) as mon:
                task_status.started()
                async for r in mon:
                    n, p, d, *m = r
                    p = pl.long(n, p)
                    m = MsgMeta.restore(m)
                    if self._chk(p, m):
                        await qw.send((p, d, m))
                if self.mark:
                    await qw.send(None)
        else:
            try:
                r = await self.link.d.get(self.path)
            except (KeyError, ValueError):
                task_status.started()
            else:
                task_status.started()
                p, d, m = Path(), r[0], MsgMeta.restore(r[1:])
                if self._chk(p, m):
                    await qw.send((p, d, m))
            if self.mark:
                await qw.send(None)

        await qw.aclose()
        if self._current_done is None:
            raise RuntimeError("Missing done event")
        self._current_done.set()

    async def _updates(self, qw, *, task_status):
        "get updates from MQTT"
        plen = len(self.path)  # ignores the root prefix
        async with self.link.monitor(
            Root.get() + self.path, subtree=self.subtree, retained=(self.state is NotGiven)
        ) as mon:
            task_status.started()
            async for msg in mon:
                p, d, m = Path.build(msg.topic[plen:]), msg.data, msg.meta
                if self._chk(p, m):
                    await qw.send((p, d, m))
        await qw.aclose()

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as tg:
            self._node = self.node_cls()
            self._current_done = anyio.Event()
            self._tg = tg
            qw, self._qr = anyio.create_memory_object_stream(10)
            if self.state is not True:
                await tg.start(self._updates, qw.clone())
            if self.state is True or self.state is None:
                await tg.start(self._current, qw.clone())
            else:
                self._current_done.set()
                if self.mark:
                    await qw.send(None)
            await qw.aclose()
            yield self
            tg.cancel_scope.cancel()
            self._node = None

    def __aiter__(self):
        return self

    @property
    @asynccontextmanager
    async def node(self) -> AsyncIterator[Node]:
        """
        Helper for fetching data in background.

        This is necessary if you start several watchers in a common context
        managers. If you then call :py.meth.`get_node` on one of them,
        the data from the other(s) might overwhelm the input queue.
        """
        if self._node is not None:
            raise RuntimeError("Use `await get_node()`")
        async with self:
            yield await self.get_node()

    async def get_node(self, background=True) -> Node:
        """
        Wait until fetching the static data is complete, then return the
        watched node.

        If @background is set (the default), start a background job that
        updates the node. In this case you cannot iterate the watcher any
        more.
        """
        if background:
            await self._tg.start(self._iter)
        if self._current_done is None:
            raise RuntimeError("Missing done event")
        await self._current_done.wait()
        return self._node

    async def _iter(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        task_status.started()
        qr, self._qr = self._qr, None

        while True:
            try:
                msg = await qr.receive()
            except anyio.EndOfStream:
                return
            p, d, m = msg
            self._node.set(p, d, m)

    async def __anext__(self):
        while True:
            try:
                msg = await self._qr.receive()
            except anyio.EndOfStream:
                raise StopAsyncIteration from None
            if msg is None:
                return None
            p, d, m = msg
            if self._current_done is None:
                raise RuntimeError("Missing done event")
            if self._node.set(p, d, m, force=self._current_done.is_set()):
                if self.meta:
                    return (p, d, m) if self.subtree else (d, m)
                else:
                    return (p, d) if self.subtree else d


class Walker:
    """
    A trimmed-down watcher that retrieves a possibly-partial subtree
    from our MoaT-Link server.

    This differs from `Watcher` by not tracking updates, nor keeping a node
    tree in memory.
    """

    def __init__(self, mon, meta=False):
        self.mon = mon
        self.pl = PathLongener()
        self.meta = meta

    def __aiter__(self):
        return self

    async def __anext__(self):
        msg = await anext(self.mon)
        n, p, d, *m = msg.args
        p = self.pl.long(n, p)
        if self.meta:
            m = MsgMeta.restore(m)
            return p, d, m
        else:
            return p, d
